import random

from .constants import WorkflowTag, WorkflowType
from .models import WorkerPoolDefinition, WorkflowDefinition

INFRAHUB_WORKER_POOL = WorkerPoolDefinition(name="infrahub-worker", description="Default Pool for internal tasks")

WEBHOOK_SEND = WorkflowDefinition(
    name="event-send-webhook",
    type=WorkflowType.USER,
    module="infrahub.webhook.tasks",
    function="send_webhook",
)

TRANSFORM_JINJA2_RENDER = WorkflowDefinition(
    name="transform_render_jinja2_template",
    type=WorkflowType.USER,
    module="infrahub.transformations.tasks",
    function="transform_render_jinja2_template",
)

TRANSFORM_PYTHON_RENDER = WorkflowDefinition(
    name="transform_render_python",
    type=WorkflowType.USER,
    module="infrahub.transformations.tasks",
    function="transform_python",
)

ANONYMOUS_TELEMETRY_SEND = WorkflowDefinition(
    name="anonymous_telemetry_send",
    type=WorkflowType.INTERNAL,
    cron=f"{random.randint(0, 59)} 2 * * *",
    module="infrahub.tasks.telemetry",
    function="send_telemetry_push",
)

SCHEMA_APPLY_MIGRATION = WorkflowDefinition(
    name="schema_apply_migrations",
    type=WorkflowType.CORE,
    module="infrahub.core.migrations.schema.tasks",
    function="schema_apply_migrations",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

SCHEMA_VALIDATE_MIGRATION = WorkflowDefinition(
    name="schema_validate_migrations",
    type=WorkflowType.CORE,
    module="infrahub.core.validators.tasks",
    function="schema_validate_migrations",
)

TRIGGER_ARTIFACT_DEFINITION_GENERATE = WorkflowDefinition(
    name="artifact-definition-generate",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="generate_artifact_definition",
)

TRIGGER_GENERATOR_DEFINITION_RUN = WorkflowDefinition(
    name="generator_definition_run",
    type=WorkflowType.CORE,
    module="infrahub.generators.tasks",
    function="run_generator_definition",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

IPAM_RECONCILIATION = WorkflowDefinition(
    name="ipam_reconciliation",
    type=WorkflowType.CORE,
    module="infrahub.core.ipam.tasks",
    function="ipam_reconciliation",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_GENERATOR_RUN = WorkflowDefinition(
    name="generator-run",
    type=WorkflowType.USER,
    module="infrahub.generators.tasks",
    function="run_generator",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_GENERATOR_DEFINITION_RUN = WorkflowDefinition(
    name="request_generator_definition_run",
    type=WorkflowType.CORE,
    module="infrahub.generators.tasks",
    function="request_generator_definition_run",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_ARTIFACT_GENERATE = WorkflowDefinition(
    name="artifact-generate",
    type=WorkflowType.CORE,  # NOTE need to check
    module="infrahub.git.tasks",
    function="generate_artifact",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_ARTIFACT_DEFINITION_GENERATE = WorkflowDefinition(
    name="request_artifact_definitions_generate",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="generate_request_artifact_definition",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

DIFF_UPDATE = WorkflowDefinition(
    name="diff-update",
    type=WorkflowType.CORE,
    module="infrahub.core.diff.tasks",
    function="update_diff",
)

DIFF_REFRESH = WorkflowDefinition(
    name="diff-refresh",
    type=WorkflowType.CORE,
    module="infrahub.core.diff.tasks",
    function="refresh_diff",
)

DIFF_REFRESH_ALL = WorkflowDefinition(
    name="diff-refresh-all",
    type=WorkflowType.INTERNAL,
    module="infrahub.core.diff.tasks",
    function="refresh_diff_all",
)

GIT_REPOSITORIES_SYNC = WorkflowDefinition(
    name="git_repositories_sync",
    type=WorkflowType.INTERNAL,
    cron="* * * * *",
    module="infrahub.git.tasks",
    function="sync_remote_repositories",
)

GIT_REPOSITORIES_CREATE_BRANCH = WorkflowDefinition(
    name="git-repositories-create-branch",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="create_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GIT_REPOSITORY_ADD = WorkflowDefinition(
    name="git-repository-add-read-write",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="add_git_repository",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GIT_REPOSITORY_ADD_READ_ONLY = WorkflowDefinition(
    name="git-repository-add-read-only",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="add_git_repository_read_only",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GIT_REPOSITORIES_PULL_READ_ONLY = WorkflowDefinition(
    name="git-repository-pull-read-only",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="pull_read_only",
)

GIT_REPOSITORIES_MERGE = WorkflowDefinition(
    name="git-repository-merge",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="merge_git_repository",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_REBASE = WorkflowDefinition(
    name="branch-rebase",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="rebase_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_CREATE = WorkflowDefinition(
    name="create-branch",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="create_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_MERGE = WorkflowDefinition(
    name="branch-merge",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="merge_branch",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_MERGE_MUTATION = WorkflowDefinition(
    name="merge-branch-mutation",
    type=WorkflowType.CORE,
    module="infrahub.graphql.mutations.tasks",
    function="merge_branch_mutation",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

BRANCH_DELETE = WorkflowDefinition(
    name="branch-delete",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="delete_branch",
)

BRANCH_VALIDATE = WorkflowDefinition(
    name="branch-validate",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="validate_branch",
)

BRANCH_CANCEL_PROPOSED_CHANGES = WorkflowDefinition(
    name="proposed-changes-cancel-branch",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="cancel_proposed_changes_branch",
)

PROPOSED_CHANGE_MERGE = WorkflowDefinition(
    name="proposed-change-merge",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="merge_proposed_change",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

GRAPHQL_QUERY_GROUP_UPDATE = WorkflowDefinition(
    name="graphql-query-group-update",
    type=WorkflowType.CORE,
    module="infrahub.groups.tasks",
    function="update_graphql_query_group",
)

PROCESS_COMPUTED_MACRO = WorkflowDefinition(
    name="process_computed_attribute_jinja2",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="process_jinja2",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES = WorkflowDefinition(
    name="trigger_update_jinja2_computed_attributes",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="trigger_update_jinja2_computed_attributes",
)

TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES = WorkflowDefinition(
    name="trigger_update_python_computed_attributes",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="trigger_update_python_computed_attributes",
)

COMPUTED_ATTRIBUTE_SETUP = WorkflowDefinition(
    name="computed-attribute-setup",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="computed_attribute_setup",
)

COMPUTED_ATTRIBUTE_SETUP_PYTHON = WorkflowDefinition(
    name="computed-attribute-setup-python",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="computed_attribute_setup_python",
)

COMPUTED_ATTRIBUTE_REMOVE_PYTHON = WorkflowDefinition(
    name="computed-attribute-remove-python",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="computed_attribute_remove_python",
)

UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM = WorkflowDefinition(
    name="process_computed_attribute_transform",
    type=WorkflowType.USER,
    module="infrahub.computed_attribute.tasks",
    function="process_transform",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS = WorkflowDefinition(
    name="query-computed-attribute-transform-targets",
    type=WorkflowType.CORE,
    module="infrahub.computed_attribute.tasks",
    function="query_transform_targets",
)

REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY = WorkflowDefinition(
    name="proposed-changed-data-integrity",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_data_integrity_check",
)

REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY = WorkflowDefinition(
    name="proposed-changed-schema-integrity",
    type=WorkflowType.CORE,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_schema_integrity_check",
)

REQUEST_PROPOSED_CHANGE_USER_TESTS = WorkflowDefinition(
    name="proposed-changed-user-tests",
    type=WorkflowType.USER,
    module="infrahub.proposed_change.tasks",
    function="run_proposed_change_user_tests",
)

AUTOMATION_BRANCH_ACTIONS = WorkflowDefinition(
    name="branch-actions-setup",
    type=WorkflowType.CORE,
    module="infrahub.core.branch.tasks",
    function="branch_actions_setup",
)

AUTOMATION_SCHEMA_UPDATED = WorkflowDefinition(
    name="schema-updated-setup",
    type=WorkflowType.CORE,
    module="infrahub.schema.tasks",
    function="schema_updated_setup",
)

AUTOMATION_GIT_UPDATED = WorkflowDefinition(
    name="git-commit-automation-setup",
    type=WorkflowType.CORE,
    module="infrahub.git.tasks",
    function="setup_commit_automation",
)

GIT_REPOSITORIES_DIFF_NAMES_ONLY = WorkflowDefinition(
    name="git-repository-diff-names-only",
    type=WorkflowType.INTERNAL,
    module="infrahub.git.tasks",
    function="git_repository_diff_names_only",
)

GIT_REPOSITORIES_IMPORT_OBJECTS = WorkflowDefinition(
    name="git-repository-import-object",
    type=WorkflowType.USER,
    module="infrahub.git.tasks",
    function="import_objects_from_git_repository",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_PROPOSED_CHANGE_RUN_GENERATORS = WorkflowDefinition(
    name="proposed-changed-run-generator",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="run_generators",
    tags=[WorkflowTag.DATABASE_CHANGE],
)

REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS = WorkflowDefinition(
    name="proposed-changed-repository-checks",
    type=WorkflowType.INTERNAL,
    module="infrahub.proposed_change.tasks",
    function="repository_checks",
)

WEBHOOK_CONFIGURE = WorkflowDefinition(
    name="webhook-setup-automations",
    type=WorkflowType.USER,
    module="infrahub.webhook.tasks",
    function="configure_webhooks",
)

WEBHOOK_TRIGGER = WorkflowDefinition(
    name="webhook-trigger-actions",
    type=WorkflowType.USER,
    module="infrahub.webhook.tasks",
    function="trigger_webhooks",
)

AUTOMATION_SETUP_WEBHOOK_CONFIGURATION_TRIGGER = WorkflowDefinition(
    name="webhook-setup-configuration-trigger",
    type=WorkflowType.USER,
    module="infrahub.webhook.tasks",
    function="trigger_webhook_configuration",
)


worker_pools = [INFRAHUB_WORKER_POOL]

workflows = [
    ANONYMOUS_TELEMETRY_SEND,
    AUTOMATION_BRANCH_ACTIONS,
    AUTOMATION_GIT_UPDATED,
    AUTOMATION_SCHEMA_UPDATED,
    AUTOMATION_SETUP_WEBHOOK_CONFIGURATION_TRIGGER,
    BRANCH_CANCEL_PROPOSED_CHANGES,
    BRANCH_CREATE,
    BRANCH_DELETE,
    BRANCH_MERGE,
    BRANCH_MERGE_MUTATION,
    BRANCH_REBASE,
    BRANCH_VALIDATE,
    COMPUTED_ATTRIBUTE_REMOVE_PYTHON,
    COMPUTED_ATTRIBUTE_SETUP,
    COMPUTED_ATTRIBUTE_SETUP_PYTHON,
    DIFF_REFRESH,
    DIFF_REFRESH_ALL,
    DIFF_UPDATE,
    GIT_REPOSITORIES_CREATE_BRANCH,
    GIT_REPOSITORIES_DIFF_NAMES_ONLY,
    GIT_REPOSITORIES_IMPORT_OBJECTS,
    GIT_REPOSITORIES_MERGE,
    GIT_REPOSITORIES_PULL_READ_ONLY,
    GIT_REPOSITORIES_SYNC,
    GIT_REPOSITORY_ADD,
    GIT_REPOSITORY_ADD_READ_ONLY,
    GRAPHQL_QUERY_GROUP_UPDATE,
    IPAM_RECONCILIATION,
    PROCESS_COMPUTED_MACRO,
    PROPOSED_CHANGE_MERGE,
    QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS,
    REQUEST_ARTIFACT_DEFINITION_GENERATE,
    REQUEST_ARTIFACT_GENERATE,
    REQUEST_GENERATOR_DEFINITION_RUN,
    REQUEST_GENERATOR_RUN,
    REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
    REQUEST_PROPOSED_CHANGE_RUN_GENERATORS,
    REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_USER_TESTS,
    SCHEMA_APPLY_MIGRATION,
    SCHEMA_VALIDATE_MIGRATION,
    TRANSFORM_JINJA2_RENDER,
    TRANSFORM_PYTHON_RENDER,
    TRIGGER_ARTIFACT_DEFINITION_GENERATE,
    TRIGGER_GENERATOR_DEFINITION_RUN,
    TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
    TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
    UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM,
    WEBHOOK_CONFIGURE,
    WEBHOOK_SEND,
    WEBHOOK_TRIGGER,
]

automation_setup_workflows = [
    AUTOMATION_BRANCH_ACTIONS,
    AUTOMATION_GIT_UPDATED,
    AUTOMATION_SCHEMA_UPDATED,
    AUTOMATION_SETUP_WEBHOOK_CONFIGURATION_TRIGGER,
]



from typing import Optional

import redis.asyncio as redis

from infrahub import config
from infrahub.message_bus.types import KVTTL
from infrahub.services import InfrahubServices
from infrahub.services.adapters.cache import InfrahubCache


class RedisCache(InfrahubCache):
    def __init__(self) -> None:
        self.connection = redis.Redis(
            host=config.SETTINGS.cache.address,
            port=config.SETTINGS.cache.service_port,
            db=config.SETTINGS.cache.database,
            ssl=config.SETTINGS.cache.tls_enabled,
            ssl_cert_reqs="optional" if not config.SETTINGS.cache.tls_insecure else "none",
            ssl_check_hostname=not config.SETTINGS.cache.tls_insecure,
            ssl_ca_certs=config.SETTINGS.cache.tls_ca_file,
        )

    async def initialize(self, service: InfrahubServices) -> None:
        pass

    async def delete(self, key: str) -> None:
        await self.connection.delete(key)

    async def get(self, key: str) -> Optional[str]:
        value = await self.connection.get(name=key)
        if value is not None:
            return value.decode()
        return None

    async def get_values(self, keys: list[str]) -> list[Optional[str]]:
        values = await self.connection.mget(keys=keys)
        return [value.decode() if value is not None else value for value in values]

    async def list_keys(self, filter_pattern: str) -> list[str]:
        cursor = 0
        has_remaining_keys = True
        keys = []
        while has_remaining_keys:
            cursor, scanned_keys = await self.connection.scan(cursor=cursor, match=filter_pattern, count=100)
            keys.extend(scanned_keys)
            if cursor == 0:
                has_remaining_keys = False

        return [key.decode() for key in keys]

    async def set(
        self, key: str, value: str, expires: Optional[KVTTL] = None, not_exists: bool = False
    ) -> Optional[bool]:
        return await self.connection.set(name=key, value=value, ex=expires.value if expires else None, nx=not_exists)



from __future__ import annotations

import importlib
import logging
import os
from enum import Enum
from typing import TYPE_CHECKING, Optional

import typer
from infrahub_sdk.async_typer import AsyncTyper
from prefect.testing.utilities import prefect_test_harness
from rich import print as rprint
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from infrahub import config
from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.graph.constraints import ConstraintManagerBase, ConstraintManagerMemgraph, ConstraintManagerNeo4j
from infrahub.core.graph.index import node_indexes, rel_indexes
from infrahub.core.graph.schema import GRAPH_SCHEMA
from infrahub.core.initialization import (
    create_anonymous_role,
    create_default_menu,
    create_default_roles,
    create_super_administrator_role,
    create_super_administrators_group,
    first_time_initialization,
    get_root_node,
    initialization,
    initialize_registry,
)
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.graph import get_graph_migrations
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.schema import SchemaRoot, core_models, internal_schema
from infrahub.core.schema.definitions.deprecated import deprecated_models
from infrahub.core.schema.manager import SchemaManager
from infrahub.core.utils import delete_all_nodes
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.database import DatabaseType
from infrahub.log import get_logger
from infrahub.services import InfrahubServices, services
from infrahub.services.adapters.message_bus.local import BusSimulator
from infrahub.services.adapters.workflow.local import WorkflowLocalExecution

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext
    from infrahub.database import InfrahubDatabase

app = AsyncTyper()

PERMISSIONS_AVAILABLE = ["read", "write", "admin"]


class ConstraintAction(str, Enum):
    SHOW = "show"
    ADD = "add"
    DROP = "drop"


class IndexAction(str, Enum):
    SHOW = "show"
    ADD = "add"
    DROP = "drop"


@app.callback()
def callback() -> None:
    """
    Manage the graph in the database.
    """


@app.command()
async def init(
    ctx: typer.Context,
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
) -> None:
    """Erase the content of the database and initialize it with the core schema."""

    log = get_logger()

    # --------------------------------------------------
    # CLEANUP
    #  - For now we delete everything in the database
    #   TODO, if possible try to implement this in an idempotent way
    # --------------------------------------------------

    logging.getLogger("neo4j").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    async with dbdriver.start_transaction() as db:
        log.info("Delete All Nodes")
        await delete_all_nodes(db=db)
        await first_time_initialization(db=db)

    await dbdriver.close()


@app.command()
async def load_test_data(
    ctx: typer.Context,
    config_file: str = typer.Option(
        "infrahub.toml", envvar="INFRAHUB_CONFIG", help="Location of the configuration file to use for Infrahub"
    ),
    dataset: str = "dataset01",
) -> None:
    """Load test data into the database from the `test_data` directory."""

    logging.getLogger("neo4j").setLevel(logging.ERROR)
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    async with dbdriver.start_session() as db:
        await initialization(db=db)

        log_level = "DEBUG"

        FORMAT = "%(message)s"
        logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
        logging.getLogger("infrahub")

        dataset_module = importlib.import_module(f"infrahub.test_data.{dataset}")
        await dataset_module.load_data(db=db)

    await dbdriver.close()


@app.command()
async def migrate(
    ctx: typer.Context,
    check: bool = typer.Option(False, help="Check the state of the database without applying the migrations."),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Check the current format of the internal graph and apply the necessary migrations"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)

    log = get_logger()

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    async with dbdriver.start_session() as db:
        rprint("Checking current state of the Database")

        await initialize_registry(db=db)
        root_node = await get_root_node(db=db)
        migrations = await get_graph_migrations(root=root_node)

        if not migrations:
            rprint(f"Database up-to-date (v{root_node.graph_version}), no migration to execute.")
        else:
            rprint(
                f"Database needs to be updated (v{root_node.graph_version} -> v{GRAPH_VERSION}), {len(migrations)} migrations pending"
            )

        if migrations and not check:
            for migration in migrations:
                log.debug(f"Execute Migration: {migration.name}")
                execution_result = await migration.execute(db=db)
                validation_result = None

                if execution_result.success:
                    validation_result = await migration.validate_migration(db=db)
                    if validation_result.success:
                        rprint(f"Migration: {migration.name} [green]SUCCESS[/green]")
                        root_node.graph_version = migration.minimum_version + 1
                        await root_node.save(db=db)

                if not execution_result.success or validation_result and not validation_result.success:
                    rprint(f"Migration: {migration.name} [bold red]FAILED[/bold red]")
                    for error in execution_result.errors:
                        rprint(f"  {error}")
                    if validation_result and not validation_result.success:
                        for error in validation_result.errors:
                            rprint(f"  {error}")
                    break

    await dbdriver.close()


@app.command()
async def update_core_schema(  # pylint: disable=too-many-statements
    ctx: typer.Context,
    debug: bool = typer.Option(False, help="Enable advanced logging and troubleshooting"),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Check the current format of the internal graph and apply the necessary migrations"""
    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)
    os.environ["PREFECT_SERVER_ANALYTICS_ENABLED"] = "false"

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    error_badge = "[bold red]ERROR[/bold red]"

    async with dbdriver.start_session() as db:
        with prefect_test_harness():
            # ----------------------------------------------------------
            # Initialize Schema and Registry
            # ----------------------------------------------------------
            service = InfrahubServices(
                database=db, message_bus=BusSimulator(database=db), workflow=WorkflowLocalExecution()
            )
            services.prepare(service)
            await initialize_registry(db=db)

            default_branch = registry.get_branch_from_registry(branch=registry.default_branch)

            registry.schema = SchemaManager()
            schema = SchemaRoot(**internal_schema)
            registry.schema.register_schema(schema=schema)

            # ----------------------------------------------------------
            # Load Current Schema from the database
            # ----------------------------------------------------------
            schema_default_branch = await registry.schema.load_schema_from_db(db=db, branch=default_branch)
            registry.schema.set_schema_branch(name=default_branch.name, schema=schema_default_branch)
            branch_schema = registry.schema.get_schema_branch(name=registry.default_branch)

            candidate_schema = branch_schema.duplicate()
            candidate_schema.load_schema(schema=SchemaRoot(**internal_schema))
            candidate_schema.load_schema(schema=SchemaRoot(**core_models))
            candidate_schema.load_schema(schema=SchemaRoot(**deprecated_models))
            candidate_schema.process()

            schema_diff = branch_schema.diff(other=candidate_schema)
            branch_schema.validate_node_deletions(diff=schema_diff)
            result = branch_schema.validate_update(
                other=candidate_schema, diff=schema_diff, enforce_update_support=False
            )
            if result.errors:
                rprint(f"{error_badge} | Unable to update the schema, due to failed validations")
                for error in result.errors:
                    rprint(error.to_string())
                raise typer.Exit(1)

            if not result.diff.all:
                await create_defaults(db=db)
                rprint("Core Schema Up to date, nothing to update")
                raise typer.Exit(0)

            rprint("Core Schema has diff, will need to be updated")
            if debug:
                result.diff.print()

            # ----------------------------------------------------------
            # Validate if the new schema is valid with the content of the database
            # ----------------------------------------------------------
            validate_migration_data = SchemaValidateMigrationData(
                branch=default_branch,
                schema_branch=candidate_schema,
                constraints=result.constraints,
            )
            responses = await schema_validate_migrations(message=validate_migration_data)
            error_messages = [violation.message for response in responses for violation in response.violations]
            if error_messages:
                rprint(f"{error_badge} | Unable to update the schema, due to failed validations")
                for message in error_messages:
                    rprint(message)
                raise typer.Exit(1)

            # ----------------------------------------------------------
            # Update the schema
            # ----------------------------------------------------------
            origin_schema = branch_schema.duplicate()

            # Update the internal schema
            schema_default_branch.load_schema(schema=SchemaRoot(**internal_schema))
            schema_default_branch.process()
            registry.schema.set_schema_branch(name=default_branch.name, schema=schema_default_branch)

            async with db.start_transaction() as dbt:
                await registry.schema.update_schema_branch(
                    schema=candidate_schema,
                    db=dbt,
                    branch=default_branch.name,
                    diff=result.diff,
                    limit=result.diff.all,
                    update_db=True,
                )
                default_branch.update_schema_hash()
                rprint("The Core Schema has been updated")
                if debug:
                    rprint(f"New schema hash: {default_branch.active_schema_hash.main}")
                await default_branch.save(db=dbt)

            # ----------------------------------------------------------
            # Run the migrations
            # ----------------------------------------------------------
            apply_migration_data = SchemaApplyMigrationData(
                branch=default_branch,
                new_schema=candidate_schema,
                previous_schema=origin_schema,
                migrations=result.migrations,
            )
            migration_error_msgs = await schema_apply_migrations(message=apply_migration_data)

            if migration_error_msgs:
                rprint(f"{error_badge} | Some error(s) happened while running the schema migrations")
                for message in migration_error_msgs:
                    rprint(message)
                raise typer.Exit(1)

            await create_defaults(db=db)


@app.command()
async def constraint(
    ctx: typer.Context,
    action: ConstraintAction = typer.Argument(ConstraintAction.SHOW),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Manage Database Constraints"""
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    manager: Optional[ConstraintManagerBase] = None
    if dbdriver.db_type == DatabaseType.NEO4J:
        manager = ConstraintManagerNeo4j.from_graph_schema(db=dbdriver, schema=GRAPH_SCHEMA)
    elif dbdriver.db_type == DatabaseType.MEMGRAPH:
        manager = ConstraintManagerMemgraph.from_graph_schema(db=dbdriver, schema=GRAPH_SCHEMA)
    else:
        print(f"Database type not supported : {dbdriver.db_type}")
        raise typer.Exit(1)

    if action == ConstraintAction.ADD:
        await manager.add()
    elif action == ConstraintAction.DROP:
        await manager.drop()

    constraints = await manager.list()

    console = Console()

    table = Table(title="Database Constraints")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Label")
    table.add_column("Property")

    for item in constraints:
        table.add_row(item.item_name, item.item_label, item.property)

    console.print(table)

    await dbdriver.close()


@app.command()
async def index(
    ctx: typer.Context,
    action: IndexAction = typer.Argument(IndexAction.SHOW),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    """Manage Database Indexes"""
    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)
    dbdriver.manager.index.init(nodes=node_indexes, rels=rel_indexes)

    if action == IndexAction.ADD:
        await dbdriver.manager.index.add()
    elif action == IndexAction.DROP:
        await dbdriver.manager.index.drop()

    indexes = await dbdriver.manager.index.list()

    console = Console()

    table = Table(title="Database Indexes")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Label")
    table.add_column("Property")
    table.add_column("Type")
    table.add_column("Entity Type")

    for item in indexes:
        table.add_row(
            item.name, item.label, ", ".join(item.properties), item.type.value.upper(), item.entity_type.value.upper()
        )

    console.print(table)

    await dbdriver.close()


async def create_defaults(db: InfrahubDatabase) -> None:
    """Create and assign default objects."""
    existing_permissions = await NodeManager.query(
        schema=InfrahubKind.OBJECTPERMISSION,
        db=db,
        limit=1,
    )
    if not existing_permissions:
        await setup_permissions(db=db)

    existing_menu_items = await NodeManager.query(
        schema=InfrahubKind.MENUITEM,
        db=db,
        limit=1,
    )
    if not existing_menu_items:
        await create_default_menu(db=db)


async def setup_permissions(db: InfrahubDatabase) -> None:
    existing_accounts = await NodeManager.query(
        schema=InfrahubKind.ACCOUNT,
        db=db,
        limit=1,
    )
    administrator_role = await create_super_administrator_role(db=db)
    await create_super_administrators_group(db=db, role=administrator_role, admin_accounts=existing_accounts)

    await create_default_roles(db=db)

    if config.SETTINGS.main.allow_anonymous_access:
        await create_anonymous_role(db=db)



from typing import Optional

from .attribute.choices import AttributeChoicesChecker
from .attribute.enum import AttributeEnumChecker
from .attribute.kind import AttributeKindChecker
from .attribute.length import AttributeLengthChecker
from .attribute.optional import AttributeOptionalChecker
from .attribute.regex import AttributeRegexChecker
from .attribute.unique import AttributeUniquenessChecker
from .interface import ConstraintCheckerInterface
from .node.attribute import NodeAttributeAddChecker
from .node.generate_profile import NodeGenerateProfileChecker
from .node.hierarchy import NodeHierarchyChecker
from .node.inherit_from import NodeInheritFromChecker
from .node.relationship import NodeRelationshipAddChecker
from .relationship.count import RelationshipCountChecker
from .relationship.optional import RelationshipOptionalChecker
from .relationship.peer import RelationshipPeerChecker
from .uniqueness.checker import UniquenessChecker

CONSTRAINT_VALIDATOR_MAP: dict[str, Optional[type[ConstraintCheckerInterface]]] = {
    "attribute.regex.update": AttributeRegexChecker,
    "attribute.enum.update": AttributeEnumChecker,
    "attribute.kind.update": AttributeKindChecker,
    "attribute.min_length.update": AttributeLengthChecker,
    "attribute.max_length.update": AttributeLengthChecker,
    "attribute.unique.update": AttributeUniquenessChecker,
    "attribute.optional.update": AttributeOptionalChecker,
    "attribute.choices.update": AttributeChoicesChecker,
    "relationship.peer.update": RelationshipPeerChecker,
    "relationship.cardinality.update": RelationshipCountChecker,
    "relationship.optional.update": RelationshipOptionalChecker,
    "relationship.min_count.update": RelationshipCountChecker,
    "relationship.max_count.update": RelationshipCountChecker,
    "node.inherit_from.update": NodeInheritFromChecker,
    "node.uniqueness_constraints.update": UniquenessChecker,
    "node.parent.update": NodeHierarchyChecker,
    "node.children.update": NodeHierarchyChecker,
    "node.generate_profile.update": NodeGenerateProfileChecker,
    "node.attribute.add": NodeAttributeAddChecker,
    "node.relationship.add": NodeRelationshipAddChecker,
}



from pathlib import Path

from .exceptions import FileNotValidError


def read_file(file_name: Path) -> str:
    if not file_name.is_file():
        raise FileNotValidError(name=str(file_name), message=f"{file_name} is not a valid file")
    try:
        with Path.open(file_name, encoding="utf-8") as fobj:
            return fobj.read()
    except UnicodeDecodeError as exc:
        raise FileNotValidError(name=str(file_name), message=f"Unable to read {file_name} with utf-8 encoding") from exc



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any

import ujson
from rich.progress import Progress

from ...queries import QUERY_RELATIONSHIPS
from ...schema import MainSchemaTypesAPI, NodeSchemaAPI
from ..constants import ILLEGAL_NAMESPACES
from ..exceptions import FileAlreadyExistsError, InvalidNamespaceError
from .interface import ExporterInterface

if TYPE_CHECKING:
    from rich.console import Console

    from ...client import InfrahubClient
    from ...node import InfrahubNode


class LineDelimitedJSONExporter(ExporterInterface):
    def __init__(self, client: InfrahubClient, console: Console | None = None):
        self.client = client
        self.console = console

    @contextmanager
    def wrapped_task_output(self, start: str, end: str = "[green]done") -> Generator:
        if self.console:
            self.console.print(f"{start}", end="...")
        yield
        if self.console:
            self.console.print(f"{end}")

    def identify_many_to_many_relationships(
        self, node_schema_map: dict[str, MainSchemaTypesAPI]
    ) -> dict[tuple[str, str], str]:
        # Identify many to many relationships by src/dst couples
        many_relationship_identifiers: dict[tuple[str, str], str] = {}

        for node_schema in node_schema_map.values():
            for relationship in node_schema.relationships:
                if (
                    relationship.cardinality != "many"
                    or not relationship.optional
                    or not relationship.identifier
                    or relationship.peer not in node_schema_map
                ):
                    continue
                for peer_relationship in node_schema_map[relationship.peer].relationships:
                    if peer_relationship.cardinality != "many" or peer_relationship.peer != node_schema.kind:
                        continue

                    forward = many_relationship_identifiers.get((node_schema.kind, relationship.peer))
                    backward = many_relationship_identifiers.get((relationship.peer, node_schema.kind))

                    # Record the relationship only if it's not known in one way or another
                    if not forward and not backward:
                        many_relationship_identifiers[node_schema.kind, relationship.peer] = relationship.identifier

        return many_relationship_identifiers

    async def retrieve_many_to_many_relationships(
        self, node_schema_map: dict[str, MainSchemaTypesAPI], branch: str
    ) -> list[dict[str, Any]]:
        has_remaining_items = True
        page_number = 1
        page_size = 50

        many_relationship_identifiers = list(self.identify_many_to_many_relationships(node_schema_map).values())
        many_relationships: list[dict[str, Any]] = []

        if not many_relationship_identifiers:
            return []

        while has_remaining_items:
            offset = (page_number - 1) * page_size

            response = await self.client.execute_graphql(
                QUERY_RELATIONSHIPS,
                variables={
                    "offset": offset,
                    "limit": page_size,
                    "relationship_identifiers": many_relationship_identifiers,
                },
                branch_name=branch,
                tracker=f"query-relationships-page{page_number}",
            )
            many_relationships.extend(response["Relationship"]["edges"])

            remaining_items = response["Relationship"]["count"] - (offset + page_size)
            if remaining_items <= 0:
                has_remaining_items = False
            page_number += 1

        return many_relationships

    # FIXME: Split in smaller functions
    async def export(  # pylint: disable=too-many-branches
        self, export_directory: Path, namespaces: list[str], branch: str, exclude: list[str] | None = None
    ) -> None:
        illegal_namespaces = set(ILLEGAL_NAMESPACES)
        node_file = export_directory / "nodes.json"
        relationship_file = export_directory / "relationships.json"

        for f in (node_file, relationship_file):
            if f.exists():
                raise FileAlreadyExistsError(f"{f.resolve()} already exists")
        if set(namespaces) & illegal_namespaces:
            raise InvalidNamespaceError(f"namespaces cannot include {illegal_namespaces}")

        with self.wrapped_task_output("Retrieving schema to export"):
            node_schema_map = await self.client.schema.all(branch=branch, namespaces=namespaces)
            node_schema_map = {
                kind: schema
                for kind, schema in node_schema_map.items()
                if isinstance(schema, NodeSchemaAPI)
                and schema.namespace not in illegal_namespaces
                and (not exclude or kind not in exclude)
            }
            retrieved_namespaces = {node_schema.namespace for node_schema in node_schema_map.values()}

        if namespaces:
            invalid_namespaces = [ns for ns in namespaces if ns not in retrieved_namespaces]
            if invalid_namespaces:
                raise InvalidNamespaceError(f"these namespaces do not exist on branch {branch}: {invalid_namespaces}")

        with self.wrapped_task_output("Retrieving many-to-many relationships"):
            many_relationships = await self.retrieve_many_to_many_relationships(node_schema_map, branch)

        schema_batch = await self.client.create_batch()
        for node_schema in node_schema_map.values():
            schema_batch.add(node_schema.kind, task=self.client.all, branch=branch)

        all_nodes: list[InfrahubNode] = []
        if self.console:
            progress = Progress()
            progress.start()
            progress_task = progress.add_task("Retrieving nodes...", total=schema_batch.num_tasks)
        async for _, schema_nodes in schema_batch.execute():
            all_nodes.extend(schema_nodes)
            if self.console:
                progress.update(progress_task, advance=1)
        if self.console:
            progress.stop()

        with self.wrapped_task_output("Writing export"):
            json_lines = [
                ujson.dumps(
                    {
                        "id": n.id,
                        "kind": n.get_kind(),
                        "graphql_json": ujson.dumps(n.get_raw_graphql_data()),
                    }
                )
                for n in all_nodes
            ]
            file_content = "\n".join(json_lines)

            if not export_directory.exists():
                export_directory.mkdir()

            node_file.write_text(file_content)
            relationship_file.write_text(ujson.dumps(many_relationships))

        if self.console:
            self.console.print(f"Export directory - {export_directory}")



from __future__ import annotations

import logging

from prefect import flow, task
from pydantic import BaseModel

from infrahub.workflows.models import WorkflowDefinition


class DummyInput(BaseModel):
    firstname: str
    lastname: str


class DummyOutput(BaseModel):
    full_name: str


DUMMY_FLOW = WorkflowDefinition(
    name="dummy_flow",
    module="infrahub.tasks.dummy",
    function="dummy_flow",
)

DUMMY_FLOW_BROKEN = WorkflowDefinition(
    name="dummy_flow_broken",
    module="infrahub.tasks.dummy",
    function="dummy_flow_broken",
)


@task
async def aggregate_name(firstname: str, lastname: str) -> str:
    return f"{firstname}, {lastname}"


@flow(name="dummy-flow", persist_result=True)
async def dummy_flow(data: DummyInput) -> DummyOutput:
    logging.getLogger("infrahub.tasks").info("Log in the flow")
    return DummyOutput(full_name=await aggregate_name(firstname=data.firstname, lastname=data.lastname))


@flow(name="dummy-flow-broken", persist_result=True)
async def dummy_flow_broken(data: DummyInput) -> DummyOutput:
    response = await aggregate_name(firstname=data.firstname, lastname=data.lastname)
    return DummyOutput(not_valid=response)  # type: ignore[call-arg]



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from infrahub.core.ipam.size import get_prefix_space
from infrahub.core.ipam.utilization import PrefixUtilizationGetter
from infrahub.core.manager import NodeManager

from . import Node

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class BuiltinIPPrefix(Node):
    async def to_graphql(
        self,
        db: InfrahubDatabase,
        fields: Optional[dict] = None,
        related_node_ids: Optional[set] = None,
        filter_sensitive: bool = False,
        permissions: Optional[dict] = None,
    ) -> dict:
        response = await super().to_graphql(
            db, fields=fields, related_node_ids=related_node_ids, filter_sensitive=filter_sensitive
        )

        if fields:
            for read_only_attr in ["netmask", "hostmask", "network_address", "broadcast_address"]:
                if read_only_attr in fields:
                    response[read_only_attr] = {"value": getattr(self.prefix, read_only_attr)}  # type: ignore[attr-defined,has-type]

            if "utilization" in fields:
                if self.member_type.id is None or self.prefix.id is None:  # type: ignore[has-type]
                    retrieved = await NodeManager.get_one(
                        db=db, branch=self._branch, id=self.id, fields={"member_type": None, "prefix": None}
                    )
                    self.member_type = retrieved.member_type  # type: ignore[union-attr]  # pylint: disable=attribute-defined-outside-init
                    self.prefix = retrieved.prefix  # type: ignore[union-attr]  # pylint: disable=attribute-defined-outside-init
                utilization_getter = PrefixUtilizationGetter(db=db, ip_prefixes=[self])
                utilization = await utilization_getter.get_use_percentage(
                    ip_prefixes=[self], branch_names=[self._branch.name]
                )
                response["utilization"] = {"value": int(utilization)}

        return response

    async def get_resource_weight(self, db: InfrahubDatabase) -> int:
        member_type = self.member_type.value  # type: ignore[has-type]  # pylint: disable=access-member-before-definition
        prefixlen = self.prefix.prefixlen  # type: ignore[has-type]  # pylint: disable=access-member-before-definition
        if member_type is None or prefixlen is None:
            retrieved = await NodeManager.get_one(
                db=db, branch=self._branch, id=self.id, fields={"member_type": None, "prefix": None}
            )
            self.member_type = retrieved.member_type  # type: ignore[union-attr]  # pylint: disable=attribute-defined-outside-init
            self.prefix = retrieved.prefix  # type: ignore[union-attr]  # pylint: disable=attribute-defined-outside-init
        return get_prefix_space(self)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Interface
from graphene.types.interface import InterfaceOptions

from infrahub.graphql.constants import KIND_GRAPHQL_FIELD_NAME

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext
    from infrahub.graphql.types import InfrahubObject


class InfrahubInterfaceOptions(InterfaceOptions):
    schema = None


class InfrahubInterface(Interface):
    @classmethod
    def resolve_type(cls, instance: dict[str, Any], info: GraphQLResolveInfo) -> InfrahubObject:
        context: GraphqlContext = info.context
        if KIND_GRAPHQL_FIELD_NAME in instance:
            return context.types[instance[KIND_GRAPHQL_FIELD_NAME]]

        raise ValueError("Unable to identify the type of the instance.")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind, RelationshipStatus
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreNumberPool
    from infrahub.database import InfrahubDatabase


class IPAddressPoolGetIdentifiers(Query):
    name = "ipaddresspool_get_identifiers"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        allocated: list[str],
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.addresses = allocated

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["addresses"] = self.addresses

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })-[reservation:IS_RESERVED]->(allocated:BuiltinIPAddress)
        WHERE allocated.uuid in $addresses
        """ % {"ipaddress_pool": InfrahubKind.IPADDRESSPOOL}
        self.add_to_query(query)
        self.return_labels = ["allocated", "reservation"]


class IPAddressPoolGetReserved(Query):
    name = "ipaddresspool_get_reserved"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["identifier"] = self.identifier

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })-[rel:IS_RESERVED]->(address:BuiltinIPAddress)
        WHERE rel.identifier = $identifier
        """ % {"ipaddress_pool": InfrahubKind.IPADDRESSPOOL}
        self.add_to_query(query)
        self.return_labels = ["address"]


class IPAddressPoolSetReserved(Query):
    name = "ipaddresspool_set_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        pool_id: str,
        address_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.address_id = address_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["address_id"] = self.address_id
        self.params["identifier"] = self.identifier

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.identifier,
        }

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })
        MATCH (address:Node { uuid: $address_id })
        CREATE (pool)-[rel:IS_RESERVED $rel_prop]->(address)
        """ % {"ipaddress_pool": InfrahubKind.IPADDRESSPOOL}

        self.add_to_query(query)
        self.return_labels = ["pool", "rel", "address"]


class NumberPoolGetAllocated(Query):
    name = "numberpool_get_allocated"
    type = QueryType.READ

    def __init__(
        self,
        pool: CoreNumberPool,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool = pool

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["node_attribute"] = self.pool.node_attribute.value
        self.params["start_range"] = self.pool.start_range.value
        self.params["end_range"] = self.pool.end_range.value

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        query = """
        MATCH (n:%(node)s)-[ha:HAS_ATTRIBUTE]-(a:Attribute {name: $node_attribute})-[hv:HAS_VALUE]-(av:AttributeValue)
        MATCH (a)-[hs:HAS_SOURCE]-(pool:%(number_pool_kind)s)
        WHERE
            av.value >= $start_range and av.value <= $end_range
            AND all(r in [ha, hv, hs] WHERE (%(branch_filter)s))
            AND ha.status = "active"
            AND hv.status = "active"
            AND hs.status = "active"
        """ % {
            "node": self.pool.node.value,
            "number_pool_kind": InfrahubKind.NUMBERPOOL,
            "branch_filter": branch_filter,
        }
        self.add_to_query(query)

        self.return_labels = ["n.uuid as id", "hv.branch as branch", "av.value as value"]
        self.order_by = ["av.value"]


class NumberPoolGetReserved(Query):
    name = "numberpool_get_reserved"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["identifier"] = self.identifier

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )

        self.params.update(branch_params)

        query = """
        MATCH (pool:%(number_pool)s { uuid: $pool_id })-[r:IS_RESERVED]->(reservation:AttributeValue)
        WHERE
            r.identifier = $identifier
            AND
            %(branch_filter)s
        """ % {"branch_filter": branch_filter, "number_pool": InfrahubKind.NUMBERPOOL}
        self.add_to_query(query)
        self.return_labels = ["reservation.value"]

    def get_reservation(self) -> int | None:
        result = self.get_result()
        if result:
            return result.get_as_optional_type("reservation.value", return_type=int)
        return None


class NumberPoolGetUsed(Query):
    name = "number_pool_get_used"
    type = QueryType.READ

    def __init__(
        self,
        pool: CoreNumberPool,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool = pool

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool.get_id()
        self.params["start_range"] = self.pool.start_range.value
        self.params["end_range"] = self.pool.end_range.value

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )

        self.params.update(branch_params)
        self.params["attribute_name"] = self.pool.node_attribute.value

        query = """
        MATCH (pool:%(number_pool)s { uuid: $pool_id })
        CALL {
            WITH pool
            MATCH (pool)-[res:IS_RESERVED]->(av:AttributeValue)<-[hv:HAS_VALUE]-(attr:Attribute)
            WHERE
                attr.name = $attribute_name
                AND
                toInteger(av.value) >= $start_range and toInteger(av.value) <= $end_range
                AND
                all(r in [res, hv] WHERE (%(branch_filter)s))
            RETURN av, (res.status = "active" AND hv.status = "active") AS is_active
        }
        WITH av, is_active
        WHERE is_active = TRUE
        """ % {
            "branch_filter": branch_filter,
            "number_pool": InfrahubKind.NUMBERPOOL,
        }
        self.add_to_query(query)
        self.return_labels = ["av.value"]
        self.order_by = ["av.value"]


class NumberPoolSetReserved(Query):
    name = "numberpool_set_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        pool_id: str,
        reserved: int,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.reserved = reserved
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["reserved"] = self.reserved
        self.params["identifier"] = self.identifier

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.identifier,
        }

        query = """
        MATCH (pool:%(number_pool)s { uuid: $pool_id })
        MERGE (value:AttributeValue { value: $reserved, is_default: false })
        CREATE (pool)-[rel:IS_RESERVED $rel_prop]->(value)
        """ % {"number_pool": InfrahubKind.NUMBERPOOL}

        self.add_to_query(query)
        self.return_labels = ["value"]


class PrefixPoolGetIdentifiers(Query):
    name = "prefixpool_get_identifiers"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        allocated: list[str],
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.prefixes = allocated

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["prefixes"] = self.prefixes

        query = """
        MATCH (pool:%(ipaddress_pool)s { uuid: $pool_id })-[reservation:IS_RESERVED]->(allocated:BuiltinIPPrefix)
        WHERE allocated.uuid in $prefixes
        """ % {"ipaddress_pool": InfrahubKind.IPPREFIXPOOL}
        self.add_to_query(query)
        self.return_labels = ["allocated", "reservation"]


class PrefixPoolGetReserved(Query):
    name = "prefixpool_get_reserved"
    type = QueryType.READ

    def __init__(
        self,
        pool_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["identifier"] = self.identifier

        query = """
        MATCH (pool:%(prefix_pool)s { uuid: $pool_id })-[rel:IS_RESERVED]->(prefix:BuiltinIPPrefix)
        WHERE rel.identifier = $identifier
        """ % {"prefix_pool": InfrahubKind.IPPREFIXPOOL}
        self.add_to_query(query)
        self.return_labels = ["prefix"]


class PrefixPoolSetReserved(Query):
    name = "prefixpool_set_reserved"
    type = QueryType.WRITE

    def __init__(
        self,
        pool_id: str,
        prefix_id: str,
        identifier: str,
        **kwargs: dict[str, Any],
    ) -> None:
        self.pool_id = pool_id
        self.prefix_id = prefix_id
        self.identifier = identifier

        super().__init__(**kwargs)  # type: ignore[arg-type]

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        self.params["pool_id"] = self.pool_id
        self.params["prefix_id"] = self.prefix_id
        self.params["identifier"] = self.identifier

        global_branch = registry.get_global_branch()
        self.params["rel_prop"] = {
            "branch": global_branch.name,
            "branch_level": global_branch.hierarchy_level,
            "status": RelationshipStatus.ACTIVE.value,
            "from": self.at.to_string(),
            "identifier": self.identifier,
        }

        query = """
        MATCH (pool:%(prefix_pool)s { uuid: $pool_id })
        MATCH (prefix:Node { uuid: $prefix_id })
        CREATE (pool)-[rel:IS_RESERVED $rel_prop]->(prefix)
        """ % {"prefix_pool": InfrahubKind.IPPREFIXPOOL}

        self.add_to_query(query)
        self.return_labels = ["pool", "rel", "prefix"]



from typing import Any, Optional

from pydantic import ConfigDict, Field, model_serializer

from infrahub.core.constants import ComputedAttributeKind
from infrahub.core.models import HashableModel


class ComputedAttribute(HashableModel):
    kind: ComputedAttributeKind
    jinja2_template: Optional[str] = Field(
        default=None, description="The Jinja2 template in string format, required when assignment_type=jinja2"
    )
    transform: Optional[str] = Field(
        default=None, description="The Python Transform name or ID, required when assignment_type=transform"
    )

    @model_serializer
    def ser_model(self) -> dict[str, Any]:
        return {"kind": self.kind.value, "jinja2_template": self.jinja2_template, "transform": self.transform}

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "allOf": [
                {
                    "if": {"properties": {"kind": {"const": "Jinja2"}}},
                    "then": {
                        "required": ["jinja2_template"],
                        "properties": {
                            "jinja2_template": {
                                "type": "string",
                                "minLength": 1,
                            }
                        },
                    },
                },
                {
                    "if": {"properties": {"kind": {"const": "TransformPython"}}},
                    "then": {
                        "required": ["transform"],
                        "properties": {
                            "transform": {
                                "type": "string",
                                "minLength": 1,
                            }
                        },
                    },
                },
            ]
        },
    )



from prefect import flow

from infrahub.core import registry
from infrahub.core.diff.model.path import BranchTrackingId
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.dependencies.registry import get_component_registry
from infrahub.log import get_logger
from infrahub.message_bus import InfrahubMessage, messages
from infrahub.services import InfrahubServices
from infrahub.workflows.catalogue import (
    DIFF_UPDATE,
    TRIGGER_ARTIFACT_DEFINITION_GENERATE,
    TRIGGER_GENERATOR_DEFINITION_RUN,
)

log = get_logger()


@flow(name="branch-event-merge")
async def merge(message: messages.EventBranchMerge, service: InfrahubServices) -> None:
    async with service.database.start_session() as db:
        log.info("Branch merged", source_branch=message.source_branch, target_branch=message.target_branch)

        events: list[InfrahubMessage] = [
            messages.RefreshRegistryBranches(),
        ]
        component_registry = get_component_registry()
        default_branch = registry.get_branch_from_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=default_branch)
        # send diff update requests for every branch-tracking diff
        branch_diff_roots = await diff_repository.get_roots_metadata(base_branch_names=[message.target_branch])

        await service.workflow.submit_workflow(
            workflow=TRIGGER_ARTIFACT_DEFINITION_GENERATE,
            parameters={"branch": message.target_branch},
        )

        await service.workflow.submit_workflow(
            workflow=TRIGGER_GENERATOR_DEFINITION_RUN,
            parameters={"branch": message.target_branch},
        )

        for diff_root in branch_diff_roots:
            if (
                diff_root.base_branch_name != diff_root.diff_branch_name
                and diff_root.tracking_id
                and isinstance(diff_root.tracking_id, BranchTrackingId)
            ):
                request_diff_update_model = RequestDiffUpdate(branch_name=diff_root.diff_branch_name)
                await service.workflow.submit_workflow(
                    workflow=DIFF_UPDATE, parameters={"model": request_diff_update_model}
                )

        for event in events:
            event.assign_meta(parent=message)
            await service.send(message=event)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, overload

from prefect.client.schemas.objects import StateType
from prefect.deployments import run_deployment

from infrahub.workflows.initialization import setup_task_manager
from infrahub.workflows.models import WorkflowInfo

from . import InfrahubWorkflow, Return

if TYPE_CHECKING:
    from prefect.client.schemas.objects import FlowRun

    from infrahub.services import InfrahubServices
    from infrahub.workflows.models import WorkflowDefinition


class WorkflowWorkerExecution(InfrahubWorkflow):
    async def initialize(self, service: InfrahubServices) -> None:
        """Initialize the Workflow engine"""

        if await service.component.is_primary_api():
            await setup_task_manager()

    @overload
    async def execute_workflow(
        self,
        workflow: WorkflowDefinition,
        expected_return: type[Return],
        parameters: dict[str, Any] | None = ...,
        tags: list[str] | None = ...,
    ) -> Return: ...

    @overload
    async def execute_workflow(
        self,
        workflow: WorkflowDefinition,
        expected_return: None = ...,
        parameters: dict[str, Any] | None = ...,
        tags: list[str] | None = ...,
    ) -> Any: ...

    async def execute_workflow(
        self,
        workflow: WorkflowDefinition,
        expected_return: type[Return] | None = None,
        parameters: dict[str, Any] | None = None,
        tags: list[str] | None = None,
    ) -> Any:
        response: FlowRun = await run_deployment(
            name=workflow.full_name, poll_interval=1, parameters=parameters or {}, tags=tags
        )  # type: ignore[return-value, misc]
        if not response.state:
            raise RuntimeError("Unable to read state from the response")

        if response.state.type == StateType.CRASHED:
            raise RuntimeError(response.state.message)

        return await response.state.result(raise_on_failure=True, fetch=True)  # type: ignore[call-overload]

    async def submit_workflow(
        self, workflow: WorkflowDefinition, parameters: dict[str, Any] | None = None, tags: list[str] | None = None
    ) -> WorkflowInfo:
        flow_run = await run_deployment(name=workflow.full_name, timeout=0, parameters=parameters or {}, tags=tags)  # type: ignore[return-value, misc]
        return WorkflowInfo.from_flow(flow_run=flow_run)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union
from uuid import UUID

from pydantic import BaseModel

from infrahub.core.constants.schema import FlagProperty, NodeProperty
from infrahub.core.registry import registry

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class ValuePropertyData(BaseModel):
    name: str
    type: str
    value: str


class NodePropertyData(BaseModel):
    name: str
    peer_id: str


class FlagPropertyMixin:
    _flag_properties: list[str] = [v.value for v in FlagProperty]

    is_visible = True
    is_protected = False

    def _init_flag_property_mixin(self, kwargs: Optional[dict] = None) -> None:
        if not kwargs:
            return

        for flag in self._flag_properties:
            if flag in kwargs.keys():
                setattr(self, flag, kwargs.get(flag))


class NodePropertyMixin:
    _node_properties: list[str] = [v.value for v in NodeProperty]

    branch: Branch
    at: Timestamp

    def _init_node_property_mixin(self, kwargs: Optional[dict] = None) -> None:
        for node in self._node_properties:
            setattr(self, f"_{node}", None)
            setattr(self, f"{node}_id", None)

        if not kwargs:
            return

        for node in self._node_properties:
            if node in kwargs.keys():
                setattr(self, node, kwargs.get(node))
            if f"{node}_id" in kwargs.keys():
                setattr(self, f"{node}_id", kwargs.get(f"{node}_id"))

    @property
    def source(self) -> Node:
        return self._get_node_property_from_cache(name="source")

    @source.setter
    def source(self, value: Union[str, Node, UUID]) -> None:
        self._set_node_property(name="source", value=value)

    @property
    def owner(self) -> Node:
        return self._get_node_property_from_cache(name="owner")

    @owner.setter
    def owner(self, value: Optional[Union[str, Node, UUID]]) -> None:
        self._set_node_property(name="owner", value=value)

    def clear_owner(self) -> None:
        self._set_node_property(name="owner", value=None)

    async def get_source(self, db: InfrahubDatabase) -> Optional[Node]:
        return await self._get_node_property(name="source", db=db)

    def clear_source(self) -> None:
        self._set_node_property(name="source", value=None)

    def set_source(self, value: Union[str, Node, UUID]) -> None:
        self._set_node_property(name="source", value=value)

    async def get_owner(self, db: InfrahubDatabase) -> Optional[Node]:
        return await self._get_node_property(name="owner", db=db)

    def set_owner(self, value: Union[str, Node, UUID]) -> None:
        self._set_node_property(name="owner", value=value)

    def _get_node_property_from_cache(self, name: str) -> Node:
        """Return the node attribute if it's already present locally,
        Otherwise raise an exception
        """
        item = getattr(self, f"_{name}", None)
        if not item:
            raise LookupError(
                f"The property {name} is not present locally, you must retrive it with get_{name}() instead."
            )

        return item

    async def _get_node_property(self, db: InfrahubDatabase, name: str) -> Optional[Node]:
        """Return the node attribute.
        If the node is already present in cache, serve from the cache
        If the node is not present, query it on the fly using the node_id
        """
        if getattr(self, f"_{name}") is None:
            await self._retrieve_node_property(db=db, name=name)

        return getattr(self, f"_{name}", None)

    def _set_node_property(self, name: str, value: Optional[Union[str, Node, UUID]]) -> None:
        """Set the value of the node_property.
        If the value is a string, we assume it's an ID and we'll save it to query it later (if needed)
        If the value is a Node, we save the node and we extract the ID
        if the value is None, we just initialize the 2 variables."""

        if isinstance(value, (str, UUID)):
            setattr(self, f"{name}_id", value)
            setattr(self, f"_{name}", None)
        elif isinstance(value, dict) and "id" in value:
            setattr(self, f"{name}_id", value["id"])
            setattr(self, f"_{name}", None)
        elif hasattr(value, "_schema"):
            setattr(self, f"_{name}", value)
            setattr(self, f"{name}_id", getattr(value, "id", None))
        elif value is None:
            setattr(self, f"_{name}", None)
            setattr(self, f"{name}_id", None)
        else:
            raise ValueError("Unable to process the node property")

    async def _retrieve_node_property(self, db: InfrahubDatabase, name: str) -> None:
        """Query the node associated with this node_property from the database."""

        node = await registry.manager.get_one(db=db, id=getattr(self, f"{name}_id"), branch=self.branch, at=self.at)
        setattr(self, f"_{name}", node)
        if node:
            setattr(self, f"{name}_id", node.id)



from enum import Enum

from infrahub.core.constants import BranchConflictKeep, InfrahubKind
from infrahub.core.integrity.object_conflict.conflict_recorder import ObjectConflictValidatorRecorder
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import SchemaNotFoundError
from infrahub.proposed_change.constants import ProposedChangeState

from .conflicts_extractor import DiffConflictsExtractor
from .model.diff import DataConflict
from .model.path import ConflictSelection, EnrichedDiffConflict, EnrichedDiffRoot, EnrichedDiffRootMetadata
from .repository.repository import DiffRepository


class DiffDataCheckSynchronizer:
    def __init__(
        self,
        db: InfrahubDatabase,
        conflicts_extractor: DiffConflictsExtractor,
        conflict_recorder: ObjectConflictValidatorRecorder,
        diff_repository: DiffRepository,
    ):
        self.db = db
        self.conflicts_extractor = conflicts_extractor
        self.conflict_recorder = conflict_recorder
        self.diff_repository = diff_repository
        self._enriched_conflicts_map: dict[str, EnrichedDiffConflict] | None = None
        self._data_conflicts: list[DataConflict] | None = None

    def _get_enriched_conflicts_map(self, enriched_diff: EnrichedDiffRoot) -> dict[str, EnrichedDiffConflict]:
        if self._enriched_conflicts_map is None:
            self._enriched_conflicts_map = enriched_diff.get_all_conflicts()
        return self._enriched_conflicts_map

    async def _get_data_conflicts(self, enriched_diff: EnrichedDiffRoot) -> list[DataConflict]:
        if self._data_conflicts is None:
            self._data_conflicts = await self.conflicts_extractor.get_data_conflicts(enriched_diff_root=enriched_diff)
        return self._data_conflicts

    async def synchronize(self, enriched_diff: EnrichedDiffRoot | EnrichedDiffRootMetadata) -> list[Node]:
        self._enriched_conflicts_map = None
        self._data_conflicts = None
        try:
            proposed_changes = await NodeManager.query(
                db=self.db,
                schema=InfrahubKind.PROPOSEDCHANGE,
                filters={"source_branch": enriched_diff.diff_branch_name, "state": ProposedChangeState.OPEN},
            )
        except SchemaNotFoundError:
            # if the CoreProposedChange schema does not exist, then there's nothing to do
            proposed_changes = []
        if not proposed_changes:
            return []
        all_data_checks = []
        for pc in proposed_changes:
            # if the enriched_diff is EnrichedDiffRootMetadata, then it has no new data in it
            if not isinstance(enriched_diff, EnrichedDiffRoot):
                has_validator = bool(await self.conflict_recorder.get_validator(proposed_change=pc))
                # if this pc does not have a validator, then it is a new ProposedChange
                if has_validator:
                    continue
                # if this is a new ProposedChange, we need to hydrate then EnrichedDiffRoot so that we can get the conflicts from it
                enriched_diff = await self.diff_repository.get_one(
                    diff_branch_name=enriched_diff.diff_branch_name, diff_id=enriched_diff.uuid
                )

            data_conflicts = await self._get_data_conflicts(enriched_diff=enriched_diff)
            enriched_conflicts_map = self._get_enriched_conflicts_map(enriched_diff=enriched_diff)
            core_data_checks = await self.conflict_recorder.record_conflicts(
                proposed_change_id=pc.get_id(), conflicts=data_conflicts
            )
            all_data_checks.extend(core_data_checks)
            core_data_checks_by_id = {cdc.enriched_conflict_id.value: cdc for cdc in core_data_checks}  # type: ignore[attr-defined]
            enriched_conflicts_by_id = {ec.uuid: ec for ec in enriched_conflicts_map.values()}
            for conflict_id, core_data_check in core_data_checks_by_id.items():
                enriched_conflict = enriched_conflicts_by_id.get(conflict_id)
                if not enriched_conflict:
                    continue
                expected_keep_branch = self._get_keep_branch_for_enriched_conflict(enriched_conflict=enriched_conflict)
                expected_keep_branch_value = (
                    expected_keep_branch.value if isinstance(expected_keep_branch, Enum) else expected_keep_branch
                )
                if core_data_check.keep_branch.value != expected_keep_branch_value:  # type: ignore[attr-defined]
                    core_data_check.keep_branch.value = expected_keep_branch_value  # type: ignore[attr-defined]
                    await core_data_check.save(db=self.db)
        return all_data_checks

    def _get_keep_branch_for_enriched_conflict(
        self, enriched_conflict: EnrichedDiffConflict
    ) -> BranchConflictKeep | None:
        if enriched_conflict.selected_branch is ConflictSelection.BASE_BRANCH:
            return BranchConflictKeep.TARGET
        if enriched_conflict.selected_branch is ConflictSelection.DIFF_BRANCH:
            return BranchConflictKeep.SOURCE
        return None



import ipaddress
from ipaddress import IPv4Interface
from typing import TYPE_CHECKING, Any, Optional

from graphene import InputObjectType, Mutation
from graphql import GraphQLResolveInfo
from typing_extensions import Self

from infrahub import lock
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.graphql.mutations.node_getter.interface import MutationNodeGetterInterface
from infrahub.lock import InfrahubMultiLock, build_object_lock_name
from infrahub.log import get_logger

from .main import InfrahubMutationMixin, InfrahubMutationOptions

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


async def validate_namespace(
    db: InfrahubDatabase,
    branch: Branch | str | None,
    data: InputObjectType,
    existing_namespace_id: Optional[str] = None,
) -> str:
    """Validate or set (if not present) the namespace to pass to the mutation and return its ID."""
    namespace_id: Optional[str] = None
    if "ip_namespace" not in data or not data["ip_namespace"]:
        namespace_id = existing_namespace_id or registry.default_ipnamespace
        data["ip_namespace"] = {"id": namespace_id}
    elif "id" in data["ip_namespace"]:
        namespace = await registry.manager.get_one(
            db=db, branch=branch, kind=InfrahubKind.IPNAMESPACE, id=data["ip_namespace"]["id"]
        )
        namespace_id = namespace.id
    else:
        raise ValidationError(
            "A valid ip_namespace must be provided or ip_namespace should be left empty in order to use the default value."
        )
    return namespace_id


class InfrahubIPNamespaceMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls,
        schema: NodeSchema,
        _meta: Optional[Any] = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ):
        if data["id"] == registry.default_ipnamespace:
            raise ValueError("Cannot delete default IPAM namespace")

        return await super().mutate_delete(info=info, data=data, branch=branch)


class InfrahubIPAddressMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls,
        schema: NodeSchema,
        _meta: Optional[Any] = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @staticmethod
    def _get_lock_name(namespace_id: str, branch: Branch) -> str | None:
        if not branch.is_default:
            # Do not lock on other branches as reconciliation will be performed at least when merging in main branch.
            return None
        return build_object_lock_name(InfrahubKind.IPADDRESS + "_" + namespace_id)

    @classmethod
    async def _mutate_create_object_and_reconcile(
        cls,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        ip_address: IPv4Interface | ipaddress.IPv6Interface,
        namespace_id: str,
    ) -> Node:
        address = await cls.mutate_create_object(data=data, db=db, branch=branch)
        reconciler = IpamReconciler(db=db, branch=branch)
        reconciled_address = await reconciler.reconcile(
            ip_value=ip_address, namespace=namespace_id, node_uuid=address.get_id()
        )
        return reconciled_address

    @classmethod
    @retry_db_transaction(name="ipaddress_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = database or context.db
        ip_address = ipaddress.ip_interface(data["address"]["value"])
        namespace_id = await validate_namespace(db=db, branch=branch, data=data)

        async with db.start_transaction() as dbt:
            if lock_name := cls._get_lock_name(namespace_id, branch):
                async with InfrahubMultiLock(lock_registry=lock.registry, locks=[lock_name]):
                    reconciled_address = await cls._mutate_create_object_and_reconcile(
                        data=data, branch=branch, db=dbt, ip_address=ip_address, namespace_id=namespace_id
                    )
            else:
                reconciled_address = await cls._mutate_create_object_and_reconcile(
                    data=data, branch=branch, db=dbt, ip_address=ip_address, namespace_id=namespace_id
                )
            result = await cls.mutate_create_to_graphql(info=info, db=dbt, obj=reconciled_address)

        return reconciled_address, result

    @classmethod
    async def _mutate_update_object_and_reconcile(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        address: Node,
        namespace_id: str,
    ) -> Node:
        address = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=address)
        reconciler = IpamReconciler(db=db, branch=branch)
        ip_address = ipaddress.ip_interface(address.address.value)
        reconciled_address = await reconciler.reconcile(
            ip_value=ip_address, node_uuid=address.get_id(), namespace=namespace_id
        )
        return reconciled_address

    @classmethod
    @retry_db_transaction(name="ipaddress_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = database or context.db

        address = node or await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        namespace = await address.ip_namespace.get_peer(db)
        namespace_id = await validate_namespace(db=db, branch=branch, data=data, existing_namespace_id=namespace.id)
        try:
            async with db.start_transaction() as dbt:
                if lock_name := cls._get_lock_name(namespace_id, branch):
                    async with InfrahubMultiLock(lock_registry=lock.registry, locks=[lock_name]):
                        reconciled_address = await cls._mutate_update_object_and_reconcile(
                            info=info, data=data, branch=branch, address=address, namespace_id=namespace_id, db=dbt
                        )
                else:
                    reconciled_address = await cls._mutate_update_object_and_reconcile(
                        info=info, data=data, branch=branch, address=address, namespace_id=namespace_id, db=dbt
                    )

                result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=reconciled_address)
        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        return address, result

    @classmethod
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getters: list[MutationNodeGetterInterface],
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self, bool]:
        context: GraphqlContext = info.context
        db = database or context.db

        await validate_namespace(db=db, branch=branch, data=data)
        prefix, result, created = await super().mutate_upsert(
            info=info, data=data, branch=branch, node_getters=node_getters, database=db
        )

        return prefix, result, created

    @classmethod
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ):
        return await super().mutate_delete(info=info, data=data, branch=branch)


class InfrahubIPPrefixMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls,
        schema: NodeSchema,
        _meta: Optional[Any] = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @staticmethod
    def _get_lock_name(namespace_id: str, branch: Branch) -> str | None:
        if not branch.is_default:
            # Do not lock on other branches as reconciliation will be performed at least when merging in main branch.
            return None
        return build_object_lock_name(InfrahubKind.IPPREFIX + "_" + namespace_id)

    @classmethod
    async def _mutate_create_object_and_reconcile(
        cls,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        namespace_id: str,
    ) -> Node:
        prefix = await cls.mutate_create_object(data=data, db=db, branch=branch)
        return await cls._reconcile_prefix(
            branch=branch, db=db, prefix=prefix, namespace_id=namespace_id, is_delete=False
        )

    @classmethod
    @retry_db_transaction(name="ipprefix_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = database or context.db
        namespace_id = await validate_namespace(db=db, branch=branch, data=data)

        async with db.start_transaction() as dbt:
            if lock_name := cls._get_lock_name(namespace_id, branch):
                async with InfrahubMultiLock(lock_registry=lock.registry, locks=[lock_name]):
                    reconciled_prefix = await cls._mutate_create_object_and_reconcile(
                        data=data, branch=branch, db=dbt, namespace_id=namespace_id
                    )
            else:
                reconciled_prefix = await cls._mutate_create_object_and_reconcile(
                    data=data, branch=branch, db=dbt, namespace_id=namespace_id
                )

            result = await cls.mutate_create_to_graphql(info=info, db=dbt, obj=reconciled_prefix)

        return reconciled_prefix, result

    @classmethod
    async def _mutate_update_object_and_reconcile(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        prefix: Node,
        namespace_id: str,
    ) -> Node:
        prefix = await cls.mutate_update_object(db=db, info=info, data=data, branch=branch, obj=prefix)
        return await cls._reconcile_prefix(
            branch=branch, db=db, prefix=prefix, namespace_id=namespace_id, is_delete=False
        )

    @classmethod
    @retry_db_transaction(name="ipprefix_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = database or context.db

        prefix = node or await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        namespace = await prefix.ip_namespace.get_peer(db)
        namespace_id = await validate_namespace(db=db, branch=branch, data=data, existing_namespace_id=namespace.id)
        try:
            async with db.start_transaction() as dbt:
                if lock_name := cls._get_lock_name(namespace_id, branch):
                    async with InfrahubMultiLock(lock_registry=lock.registry, locks=[lock_name]):
                        reconciled_prefix = await cls._mutate_update_object_and_reconcile(
                            info=info, data=data, prefix=prefix, db=dbt, namespace_id=namespace_id, branch=branch
                        )
                else:
                    reconciled_prefix = await cls._mutate_update_object_and_reconcile(
                        info=info, data=data, prefix=prefix, db=dbt, namespace_id=namespace_id, branch=branch
                    )
                result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=reconciled_prefix)
        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        return prefix, result

    @classmethod
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getters: list[MutationNodeGetterInterface],
        database: Optional[InfrahubDatabase] = None,
    ):
        context: GraphqlContext = info.context
        db = database or context.db

        await validate_namespace(db=db, branch=branch, data=data)
        prefix, result, created = await super().mutate_upsert(
            info=info, data=data, branch=branch, node_getters=node_getters, database=db
        )

        return prefix, result, created

    @classmethod
    async def _reconcile_prefix(
        cls,
        branch: Branch,
        db: InfrahubDatabase,
        prefix: Node,
        namespace_id: str,
        is_delete: bool,
    ) -> Node:
        reconciler = IpamReconciler(db=db, branch=branch)
        ip_network = ipaddress.ip_network(prefix.prefix.value)
        reconciled_prefix = await reconciler.reconcile(
            ip_value=ip_network, node_uuid=prefix.get_id(), namespace=namespace_id, is_delete=is_delete
        )
        return reconciled_prefix

    @classmethod
    @retry_db_transaction(name="ipprefix_delete")
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context
        db = context.db

        prefix = await NodeManager.get_one(data.get("id"), context.db, branch=branch, prefetch_relationships=True)
        if not prefix:
            raise NodeNotFoundError(branch, cls._meta.schema.kind, data.get("id"))

        namespace_rels = await prefix.ip_namespace.get_relationships(db=db)
        namespace_id = namespace_rels[0].peer_id
        try:
            async with context.db.start_transaction() as dbt:
                if lock_name := cls._get_lock_name(namespace_id, branch):
                    async with InfrahubMultiLock(lock_registry=lock.registry, locks=[lock_name]):
                        reconciled_prefix = await cls._reconcile_prefix(
                            branch=branch, db=dbt, prefix=prefix, namespace_id=namespace_id, is_delete=True
                        )
                else:
                    reconciled_prefix = await cls._reconcile_prefix(
                        branch=branch, db=dbt, prefix=prefix, namespace_id=namespace_id, is_delete=True
                    )

        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        ok = True

        return reconciled_prefix, cls(ok=ok)



import base64
import hashlib
import hmac
from datetime import datetime, timezone
from math import floor
from typing import Any, Optional, Union
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field

from infrahub.core.constants import InfrahubKind
from infrahub.git.repository import InfrahubReadOnlyRepository, InfrahubRepository
from infrahub.services import InfrahubServices


class SendWebhookData(BaseModel):
    """Sent a webhook to an external source."""

    webhook_id: str = Field(..., description="The unique ID of the webhook")
    event_type: str = Field(..., description="The event type")
    event_data: dict = Field(..., description="The data tied to the event")


class Webhook(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    service: InfrahubServices = Field(...)
    url: str = Field(...)
    event_type: str = Field(...)
    data: dict[str, Any] = Field(...)
    validate_certificates: bool = Field(...)
    _payload: Any = None
    _headers: Optional[dict[str, Any]] = None

    async def _prepare_payload(self) -> None:
        self._payload = {"event_type": self.event_type, "data": self.data}

    def _assign_headers(self) -> None:
        self._headers = {}

    @property
    def webhook_type(self) -> str:
        return self.__class__.__name__

    async def send(self) -> None:
        await self._prepare_payload()
        self._assign_headers()
        await self.service.http.post(url=self.url, json=self._payload, headers=self._headers)


class CustomWebhook(Webhook):
    """Custom webhook"""


class StandardWebhook(Webhook):
    shared_key: bytes = Field(...)

    def _assign_headers(self) -> None:
        message_id = f"msg_{uuid4().hex}"
        timestamp = str(floor(datetime.now(tz=timezone.utc).timestamp()))
        payload = self._payload or {}
        unsigned_data = f"{message_id}.{timestamp}.{payload}".encode()
        signature = self._sign(data=unsigned_data)

        self._headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "webhook-id": message_id,
            "webhook-timestamp": timestamp,
            "webhook-signature": f"v1,{base64.b64encode(signature).decode('utf-8')}",
        }

    def _sign(self, data: bytes) -> bytes:
        return hmac.new(key=self.shared_key, msg=data, digestmod=hashlib.sha256).digest()


class TransformWebhook(Webhook):
    repository_id: str = Field(...)
    repository_name: str = Field(...)
    repository_kind: str = Field(...)
    transform_name: str = Field(...)
    transform_class: str = Field(...)
    transform_file: str = Field(...)

    async def _prepare_payload(self) -> None:
        repo: Union[InfrahubReadOnlyRepository, InfrahubRepository]
        if self.repository_kind == InfrahubKind.READONLYREPOSITORY:
            repo = await InfrahubReadOnlyRepository.init(
                id=self.repository_id, name=self.repository_name, client=self.service.client
            )
        else:
            repo = await InfrahubRepository.init(
                id=self.repository_id, name=self.repository_name, client=self.service.client
            )

        default_branch = repo.default_branch
        commit = repo.get_commit_value(branch_name=default_branch)

        self._payload = await repo.execute_python_transform(
            branch_name=default_branch,
            commit=commit,
            location=f"{self.transform_file}::{self.transform_class}",
            data={"event_type": self.event_type, "data": self.data},
            client=self.service.client,
        )



from typing import Optional

from graphene import InputObjectType

from infrahub.core.branch import Branch
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase

from .interface import MutationNodeGetterInterface


class MutationNodeGetterById(MutationNodeGetterInterface):
    def __init__(self, db: InfrahubDatabase, node_manager: NodeManager) -> None:
        self.db = db
        self.node_manager = node_manager

    async def get_node(
        self,
        node_schema: MainSchemaTypes,
        data: InputObjectType,
        branch: Branch,
    ) -> Optional[Node]:
        node = None
        if "id" not in data:
            return node
        return await self.node_manager.get_one(id=str(data["id"]), db=self.db, branch=branch, kind=node_schema.kind)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from infrahub.core.constants import PathType, RelationshipCardinality
from infrahub.core.path import DataPath, GroupedDataPaths

from ..interface import ConstraintCheckerInterface
from ..shared import (
    RelationshipSchemaValidatorQuery,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class RelationshipCountUpdateValidatorQuery(RelationshipSchemaValidatorQuery):
    name = "relationship_constraints_count_validator"

    def __init__(
        self,
        min_count_override: Optional[int] = None,
        max_count_override: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        self.min_count_override = min_count_override
        self.max_count_override = max_count_override
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)

        self.params["relationship_id"] = self.relationship_schema.identifier
        self.params["relationship_direction"] = self.relationship_schema.direction.value
        self.params["min_count"] = (
            self.min_count_override if self.min_count_override is not None else self.relationship_schema.min_count
        )
        max_count: int | None = self.relationship_schema.max_count
        if self.max_count_override:
            max_count = self.max_count_override
        if max_count == 0:
            max_count = None
        self.params["max_count"] = max_count

        # ruff: noqa: E501
        query = """
        // get the nodes on these branches nodes
        MATCH (n:%(node_kind)s)
        CALL {
            WITH n
            MATCH path = (root:Root)<-[rroot:IS_PART_OF]-(n)
            WHERE all(r in relationships(path) WHERE %(branch_filter)s)
            RETURN path as full_path, n as active_node
            ORDER BY rroot.branch_level DESC, rroot.from DESC
            LIMIT 1
        }
        // filter to only the active nodes
        WITH full_path, active_node
        WITH full_path, active_node
        WHERE all(r in relationships(full_path) WHERE r.status = "active")
        // get the relationships using the given identifier for each node
        CALL {
            WITH active_node
            MATCH path = (active_node)-[rrel1:IS_RELATED]-(rel:Relationship { name: $relationship_id })-[rrel2:IS_RELATED]-(peer:Node)
            WHERE ($relationship_direction <> "outbound" OR (startNode(rrel1) = active_node AND startNode(rrel2) = rel))
            AND ($relationship_direction <> "inbound" OR (startNode(rrel1) = rel AND startNode(rrel2) = peer))
            AND all(
                r in relationships(path)
                WHERE (%(branch_filter)s)
            )
            RETURN
                path as relationship_path,
                active_node as start_node,
                peer as peer_node,
                rrel1.branch_level + rrel2.branch_level AS branch_level_sum,
                [rrel1.from, rrel2.from] as from_times,
                // used as tiebreaker for updated relationships that were deleted and added at the same microsecond
                reduce(active_count = 0, r in relationships(path) | active_count + (CASE r.status WHEN "active" THEN 1 ELSE 0 END)) AS active_relationship_count,
                rrel1.branch as node_branch_name
        }
        WITH
            collect([branch_level_sum, from_times, active_relationship_count, relationship_path, node_branch_name]) as enriched_paths,
            start_node,
            peer_node
        // make sure to only use the latest version of this particular path
        CALL {
            WITH enriched_paths, peer_node
            UNWIND enriched_paths as path_to_check
            RETURN path_to_check[3] as current_path, path_to_check[4] as branch_name, peer_node as current_peer
            ORDER BY
                path_to_check[0] DESC,
                path_to_check[1][1] DESC,
                path_to_check[1][0] DESC,
                path_to_check[2] DESC
            LIMIT 1
        }
        // filter to only the current active paths
        WITH collect([current_peer, current_path]) as peers_and_paths, start_node, branch_name
        CALL {
            WITH peers_and_paths
            UNWIND peers_and_paths AS peer_and_path
            WITH peer_and_path
            WHERE all(r in relationships(peer_and_path[1]) WHERE r.status = "active")
            RETURN count(peer_and_path[0]) as num_relationships_on_branch
        }
        // sum all the relationships across branches and identify the violators
        WITH collect([branch_name, num_relationships_on_branch]) as branches_and_counts, start_node
        CALL {
            WITH start_node, branches_and_counts
            WITH start_node, branches_and_counts, reduce(rel_total = 0, bnc in branches_and_counts | rel_total + bnc[1]) AS total_relationships_count
            WHERE
                (toInteger($min_count) IS NOT NULL AND total_relationships_count < toInteger($min_count))
                OR (toInteger($max_count) IS NOT NULL AND total_relationships_count > toInteger($max_count))
            RETURN start_node as violation_node, branches_and_counts as violation_branches_and_counts
        }
        // return a row for each node-branch combination with a count for that branch
        UNWIND violation_branches_and_counts as violation_branch_and_count
        """ % {"branch_filter": branch_filter, "node_kind": self.node_schema.kind}

        self.add_to_query(query)
        self.return_labels = [
            "start_node.uuid as node_uuid",
            "violation_branch_and_count[0] as branch_name",
            "violation_branch_and_count[1] as num_relationships",
        ]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            node_id = str(result.get("node_uuid"))
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("branch_name")),
                    path_type=PathType.NODE,
                    node_id=node_id,
                    field_name=self.relationship_schema.name,
                    kind=self.node_schema.kind,
                    value=result.get("num_relationships"),
                ),
                grouping_key=node_id,
            )

        return grouped_data_paths


class RelationshipCountChecker(ConstraintCheckerInterface):
    query_classes = [RelationshipCountUpdateValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Optional[Branch] = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "relationship.count.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name in (
            "relationship.min_count.update",
            "relationship.max_count.update",
            "relationship.cardinality.update",
        )

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []
        if not request.schema_path.field_name:
            raise ValueError("field_name is not defined")
        relationship_schema = request.node_schema.get_relationship(name=request.schema_path.field_name)
        min_count_override, max_count_override = None, None
        if request.constraint_name == "relationship.cardinality.update":
            if relationship_schema.cardinality == RelationshipCardinality.ONE:
                max_count_override = 1
                min_count_override = 0 if relationship_schema.optional else 1
            else:
                return grouped_data_paths_list

        elif not (relationship_schema.min_count or relationship_schema.max_count):
            return grouped_data_paths_list

        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db,
                branch=self.branch,
                node_schema=request.node_schema,
                schema_path=request.schema_path,
                min_count_override=min_count_override,
                max_count_override=max_count_override,
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



from infrahub.core.diff.enricher.hierarchy import DiffHierarchyEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffHierarchyEnricherDependency(DependencyBuilder[DiffHierarchyEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffHierarchyEnricher:
        return DiffHierarchyEnricher(db=context.db)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub import config
from infrahub.core.constants import RelationshipStatus
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class AddNodeToBranch(Query):
    name: str = "node_add_to_branch"
    insert_return: bool = False

    type: QueryType = QueryType.WRITE

    def __init__(self, node_id: int, **kwargs: Any):
        self.node_id = node_id
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        query = """
        MATCH (root:Root)
        MATCH (d) WHERE %(id_func)s(d) = $node_id
        WITH root,d
        CREATE (d)-[r:IS_PART_OF { branch: $branch, branch_level: $branch_level, from: $now, status: $status }]->(root)
        RETURN %(id_func)s(r)
        """ % {
            "id_func": db.get_id_function_name(),
        }

        self.params["node_id"] = db.to_database_id(self.node_id)
        self.params["now"] = self.at.to_string()
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["status"] = RelationshipStatus.ACTIVE.value

        self.add_to_query(query)


class DeleteBranchRelationshipsQuery(Query):
    name: str = "delete_branch_relationships"
    insert_return: bool = False

    type: QueryType = QueryType.WRITE

    def __init__(self, branch_name: str, **kwargs: Any):
        self.branch_name = branch_name
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        if config.SETTINGS.database.db_type == config.DatabaseType.MEMGRAPH:
            query = """
            MATCH p = (s)-[r]-(d)
            WHERE r.branch = $branch_name
            DELETE r
            """
        else:
            query = """
            MATCH p = (s)-[r]-(d)
            WHERE r.branch = $branch_name
            DELETE r
            WITH *
            UNWIND nodes(p) AS n
            MATCH (n)
            WHERE NOT exists((n)--())
            DELETE n
            """
        self.params["branch_name"] = self.branch_name
        self.add_to_query(query)


class GetAllBranchInternalRelationshipQuery(Query):
    name: str = "get_internal_relationship"

    type: QueryType = QueryType.READ
    insert_return: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        query = """
        MATCH p = ()-[r]-()
        WHERE r.branch = $branch_name
        RETURN DISTINCT r
        """
        self.add_to_query(query=query)
        self.params["branch_name"] = self.branch.name
        self.return_labels = ["r"]


class RebaseBranchUpdateRelationshipQuery(Query):
    name: str = "rebase_branch_update"

    type: QueryType = QueryType.WRITE

    def __init__(self, ids: list[str], **kwargs: Any) -> None:
        self.ids = ids
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        query = """
        MATCH ()-[r]->()
        WHERE %(id_func)s(r) IN $ids
        SET r.from = $at
        SET r.conflict = NULL
        """ % {
            "id_func": db.get_id_function_name(),
        }

        self.add_to_query(query=query)

        self.params["at"] = self.at.to_string()
        self.params["ids"] = [db.to_database_id(id) for id in self.ids]
        self.return_labels = [f"{db.get_id_function_name()}(r)"]


class RebaseBranchDeleteRelationshipQuery(Query):
    name: str = "rebase_branch_delete"

    type: QueryType = QueryType.WRITE
    insert_return: bool = False

    def __init__(self, ids: list[str], **kwargs: Any) -> None:
        self.ids = ids
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        if config.SETTINGS.database.db_type == config.DatabaseType.MEMGRAPH:
            query = """
            MATCH p = (s)-[r]-(d)
            WHERE %(id_func)s(r) IN $ids
            DELETE r
            """
        else:
            query = """
            MATCH p = (s)-[r]-(d)
            WHERE %(id_func)s(r) IN $ids
            DELETE r
            WITH *
            UNWIND nodes(p) AS n
            MATCH (n)
            WHERE NOT exists((n)--())
            DELETE n
            """
        query %= {
            "id_func": db.get_id_function_name(),
        }

        self.add_to_query(query=query)

        self.params["ids"] = [db.to_database_id(id) for id in self.ids]



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from typing import Optional

from pydantic import Field

from infrahub.core.schema.basenode_schema import BaseNodeSchema


class GeneratedNodeSchema(BaseNodeSchema):
    inherit_from: list[str] = Field(
        default_factory=list,
        description="List of Generic Kind that this node is inheriting from",
        json_schema_extra={"update": "validate_constraint"},
    )
    generate_profile: bool = Field(
        default=True,
        description="Indicate if a profile schema should be generated for this schema",
        json_schema_extra={"update": "validate_constraint"},
    )
    hierarchy: Optional[str] = Field(
        default=None,
        description="Internal value to track the name of the Hierarchy, must match the name of a Generic supporting hierarchical mode",
        json_schema_extra={"update": "validate_constraint"},
    )
    parent: Optional[str] = Field(
        default=None,
        description="Expected Kind for the parent node in a Hierarchy, default to the main generic defined if not defined.",
        json_schema_extra={"update": "validate_constraint"},
    )
    children: Optional[str] = Field(
        default=None,
        description="Expected Kind for the children nodes in a Hierarchy, default to the main generic defined if not defined.",
        json_schema_extra={"update": "validate_constraint"},
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import PathType
from infrahub.core.path import DataPath, GroupedDataPaths

from .shared import SchemaValidatorQuery

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class NodeNotPresentValidatorQuery(SchemaValidatorQuery):
    name: str = "node_not_present_validator"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        query = """
        MATCH (n:%(node_kind)s)
        CALL {
            WITH n
            MATCH path = (root:Root)<-[rr:IS_PART_OF]-(n)
            WHERE all(
                r in relationships(path)
                WHERE %(branch_filter)s
            )
            RETURN path as full_path, n as node, rr as root_relationship
            ORDER BY rr.branch_level DESC, rr.from DESC
            LIMIT 1
        }
        WITH full_path, node, root_relationship
        WITH full_path, node, root_relationship
        WHERE all(r in relationships(full_path) WHERE r.status = "active")
        """ % {"branch_filter": branch_filter, "node_kind": self.node_schema.kind}

        self.add_to_query(query)
        self.return_labels = ["node.uuid", "root_relationship"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("root_relationship").get("branch")),
                    path_type=PathType.NODE,
                    node_id=str(result.get("node.uuid")),
                    kind=self.node_schema.kind,
                ),
            )

        return grouped_data_paths



from typing import TYPE_CHECKING, Any

from graphene import Boolean, Field, InputField, InputObjectType, Mutation, String
from graphql import GraphQLResolveInfo
from infrahub_sdk.utils import extract_fields
from infrahub_sdk.uuidt import UUIDT
from typing_extensions import Self

from infrahub.auth import AuthType
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreAccount, CoreNode, InternalAccountToken
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import NodeNotFoundError, PermissionDeniedError

from ..models import OrderModel
from ..types import InfrahubObjectType

if TYPE_CHECKING:
    from ..initialization import GraphqlContext


# pylint: disable=unused-argument


class InfrahubAccountTokenCreateInput(InputObjectType):
    name = InputField(String(required=False), description="The name of the token")
    expiration = InputField(String(required=False), description="Timestamp when the token expires")


class InfrahubAccountTokenDeleteInput(InputObjectType):
    id = InputField(String(required=True), description="The id of the token to delete")


class InfrahubAccountUpdateSelfInput(InputObjectType):
    password = InputField(String(required=False), description="Password to use instead of the current one")
    description = InputField(String(required=False), description="Description to use instead of the current one")


class ValueType(InfrahubObjectType):
    value = String(required=True)


class InfrahubAccountTokenType(InfrahubObjectType):
    id = String(required=True)
    token = Field(ValueType)


class AccountMixin:
    @classmethod
    async def mutate(
        cls,
        root: dict,
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> Self:
        context: GraphqlContext = info.context

        if not context.account_session:
            raise ValueError("An account_session is mandatory to execute this mutation")

        if context.account_session.auth_type != AuthType.JWT:
            raise PermissionDeniedError("This operation requires authentication with a JWT token")

        results = await NodeManager.query(
            schema=CoreAccount, filters={"ids": [context.account_session.account_id]}, db=context.db
        )
        if not results:
            raise NodeNotFoundError(node_type=InfrahubKind.ACCOUNT, identifier=context.account_session.account_id)

        account = results[0]

        mutation_map = {
            "InfrahubAccountTokenCreate": cls.create_token,
            "InfrahubAccountTokenDelete": cls.delete_token,
            "InfrahubAccountSelfUpdate": cls.update_self,
        }
        response = await mutation_map[cls.__name__](db=context.db, account=account, data=data, info=info)

        # Reset the time of the query to guarantee that all resolvers executed after this point will account for the changes
        context.at = Timestamp()

        return response

    @classmethod
    @retry_db_transaction(name="account_token_create")
    async def create_token(
        cls, db: InfrahubDatabase, account: CoreNode, data: dict[str, Any], info: GraphQLResolveInfo
    ) -> Self:
        obj = await Node.init(db=db, schema=InfrahubKind.ACCOUNTTOKEN)
        token = str(UUIDT())
        await obj.new(
            db=db,
            account=account,
            token=token,
            name=data.get("name"),
            expiration=data.get("expiration"),
        )

        async with db.start_transaction() as dbt:
            await obj.save(db=dbt)

        fields = await extract_fields(info.field_nodes[0].selection_set)
        return cls(object=await obj.to_graphql(db=db, fields=fields.get("object", {})), ok=True)  # type: ignore[call-arg]

    @classmethod
    @retry_db_transaction(name="account_token_delete")
    async def delete_token(
        cls, db: InfrahubDatabase, account: CoreNode, data: dict[str, Any], info: GraphQLResolveInfo
    ) -> Self:
        token_id = str(data.get("id"))

        results = await NodeManager.query(
            schema=InternalAccountToken,
            filters={"account_ids": [account.id], "ids": [token_id]},
            db=db,
            order=OrderModel(disable=True),
        )

        if not results:
            raise NodeNotFoundError(node_type="AccountToken", identifier=token_id)

        async with db.start_transaction() as dbt:
            await results[0].delete(db=dbt)

        return cls(ok=True)  # type: ignore[call-arg]

    @classmethod
    @retry_db_transaction(name="account_update_self")
    async def update_self(
        cls, db: InfrahubDatabase, account: CoreNode, data: dict[str, Any], info: GraphQLResolveInfo
    ) -> Self:
        for field in ("password", "description"):
            if value := data.get(field):
                getattr(account, field).value = value

        async with db.start_transaction() as dbt:
            await account.save(db=dbt)

        return cls(ok=True)  # type: ignore[call-arg]


class InfrahubAccountTokenCreate(AccountMixin, Mutation):
    class Arguments:
        data = InfrahubAccountTokenCreateInput(required=True)

    ok = Boolean()
    object = Field(InfrahubAccountTokenType)


class InfrahubAccountTokenDelete(AccountMixin, Mutation):
    class Arguments:
        data = InfrahubAccountTokenDeleteInput(required=True)

    ok = Boolean()


class InfrahubAccountSelfUpdate(AccountMixin, Mutation):
    class Arguments:
        data = InfrahubAccountUpdateSelfInput(required=True)

    ok = Boolean()



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventBranchCreate(InfrahubMessage):
    """Sent a new branch is created."""

    branch: str = Field(..., description="The branch that was created")
    branch_id: str = Field(..., description="The unique ID of the branch")
    sync_with_git: bool = Field(..., description="Indicates if Infrahub should extend this branch to git.")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from graphene import InputObjectType, Mutation
from typing_extensions import Self

from infrahub.core.schema import NodeSchema
from infrahub.git.models import RequestArtifactDefinitionGenerate
from infrahub.log import get_logger
from infrahub.workflows.catalogue import REQUEST_ARTIFACT_DEFINITION_GENERATE

from .main import InfrahubMutationMixin, InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


class InfrahubArtifactDefinitionMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls,
        schema: NodeSchema,
        _meta: Optional[Any] = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context

        artifact_definition, result = await super().mutate_create(info=info, data=data, branch=branch)

        if context.service:
            model = RequestArtifactDefinitionGenerate(branch=branch.name, artifact_definition=artifact_definition.id)
            await context.service.workflow.submit_workflow(
                workflow=REQUEST_ARTIFACT_DEFINITION_GENERATE, parameters={"model": model}
            )

        return artifact_definition, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context

        artifact_definition, result = await super().mutate_update(info=info, data=data, branch=branch)

        if context.service:
            model = RequestArtifactDefinitionGenerate(branch=branch.name, artifact_definition=artifact_definition.id)
            await context.service.workflow.submit_workflow(
                workflow=REQUEST_ARTIFACT_DEFINITION_GENERATE, parameters={"model": model}
            )

        return artifact_definition, result



from typing import Any

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import (
    EnrichedDiffAttribute,
    EnrichedDiffConflict,
    EnrichedDiffNode,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffs,
    EnrichedDiffSingleRelationship,
    EnrichedNodeCreateRequest,
)


class EnrichedDiffRootsCreateQuery(Query):
    name = "enriched_roots_create"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, enriched_diffs: EnrichedDiffs, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.enriched_diffs = enriched_diffs

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        self.params = self._build_diff_root_params(enriched_diffs=self.enriched_diffs)
        query = """
UNWIND $diff_root_list AS diff_root_map
WITH diff_root_map
CALL {
    WITH diff_root_map
    MERGE (diff_root:DiffRoot {
        base_branch: diff_root_map.base_branch,
        diff_branch: diff_root_map.diff_branch,
        from_time: diff_root_map.from_time,
        to_time: diff_root_map.to_time,
        uuid: diff_root_map.uuid,
        num_added: diff_root_map.num_added,
        num_updated: diff_root_map.num_updated,
        num_removed: diff_root_map.num_removed,
        num_conflicts: diff_root_map.num_conflicts,
        contains_conflict: diff_root_map.contains_conflict
    })
    SET diff_root.tracking_id = diff_root_map.tracking_id
    RETURN diff_root
}
WITH DISTINCT diff_root AS diff_root
WITH collect(diff_root) AS diff_roots
CALL {
    WITH diff_roots
    WITH diff_roots[0] AS base_diff_node, diff_roots[1] AS branch_diff_node
    MERGE (base_diff_node)-[:DIFF_HAS_PARTNER]-(branch_diff_node)
    SET (base_diff_node).partner_uuid = (branch_diff_node).uuid
    SET (branch_diff_node).partner_uuid = (base_diff_node).uuid
}
        """
        self.add_to_query(query)

    def _build_diff_root_params(self, enriched_diffs: EnrichedDiffs) -> dict[str, Any]:
        diff_root_list: list[dict[str, Any]] = []
        for enriched_diff in (enriched_diffs.base_branch_diff, enriched_diffs.diff_branch_diff):
            diff_root_list.append(
                {
                    "base_branch": enriched_diff.base_branch_name,
                    "diff_branch": enriched_diff.diff_branch_name,
                    "from_time": enriched_diff.from_time.to_string(),
                    "to_time": enriched_diff.to_time.to_string(),
                    "uuid": enriched_diff.uuid,
                    "tracking_id": enriched_diff.tracking_id.serialize() if enriched_diff.tracking_id else None,
                    "num_added": enriched_diff.num_added,
                    "num_updated": enriched_diff.num_updated,
                    "num_removed": enriched_diff.num_removed,
                    "num_conflicts": enriched_diff.num_conflicts,
                    "contains_conflict": enriched_diff.contains_conflict,
                }
            )
        return {"diff_root_list": diff_root_list}


class EnrichedNodeBatchCreateQuery(Query):
    name = "enriched_nodes_create"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, node_create_batch: list[EnrichedNodeCreateRequest], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.node_create_batch = node_create_batch

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        self.params = self._build_node_batch_params()
        query = """
UNWIND $node_details_list AS node_details
WITH node_details.root_uuid AS root_uuid, node_details.node_map AS node_map
CALL {
    WITH root_uuid, node_map
    MATCH (diff_root {uuid: root_uuid})
    CREATE (diff_root)-[:DIFF_HAS_NODE]->(diff_node:DiffNode)
    SET diff_node = node_map.node_properties
    // -------------------------
    // add node-level conflict
    // -------------------------
    FOREACH (i in CASE WHEN node_map.conflict_params IS NOT NULL THEN [1] ELSE [] END |
        CREATE (diff_node)-[:DIFF_HAS_CONFLICT]->(diff_node_conflict:DiffConflict)
        SET diff_node_conflict = node_map.conflict_params
    )
    // -------------------------
    // add attributes for this node
    // -------------------------
    WITH diff_node, node_map
    CALL {
        WITH diff_node, node_map
        UNWIND node_map.attributes AS node_attribute
        CREATE (diff_node)-[:DIFF_HAS_ATTRIBUTE]->(diff_attribute:DiffAttribute)
        SET diff_attribute = node_attribute.node_properties
        // -------------------------
        // add properties for this attribute
        // -------------------------
        WITH diff_attribute, node_attribute
        UNWIND node_attribute.properties AS attr_property
        CREATE (diff_attribute)-[:DIFF_HAS_PROPERTY]->(diff_attr_prop:DiffProperty)
        SET diff_attr_prop = attr_property.node_properties
        // -------------------------
        // add conflict for this property
        // -------------------------
        FOREACH (i in CASE WHEN attr_property.conflict_params IS NOT NULL THEN [1] ELSE [] END |
            CREATE (diff_attr_prop)-[:DIFF_HAS_CONFLICT]->(diff_attribute_property_conflict:DiffConflict)
            SET diff_attribute_property_conflict = attr_property.conflict_params
        )
    }
    // -------------------------
    // add relationships for this node
    // -------------------------
    WITH diff_node, node_map
    CALL {
        WITH diff_node, node_map
        UNWIND node_map.relationships as node_relationship
        CREATE (diff_node)-[:DIFF_HAS_RELATIONSHIP]->(diff_relationship:DiffRelationship)
        SET diff_relationship = node_relationship.node_properties
        // -------------------------
        // add elements for this relationship group
        // -------------------------
        WITH diff_relationship, node_relationship
        UNWIND node_relationship.relationships as node_single_relationship
        CREATE (diff_relationship)-[:DIFF_HAS_ELEMENT]->(diff_relationship_element:DiffRelationshipElement)
        SET diff_relationship_element = node_single_relationship.node_properties
        // -------------------------
        // add conflict for this relationship element
        // -------------------------
        FOREACH (i in CASE WHEN node_single_relationship.conflict_params IS NOT NULL THEN [1] ELSE [] END |
            CREATE (diff_relationship_element)-[:DIFF_HAS_CONFLICT]->(diff_relationship_conflict:DiffConflict)
            SET diff_relationship_conflict = node_single_relationship.conflict_params
        )
        // -------------------------
        // add properties for this relationship element
        // -------------------------
        WITH diff_relationship_element, node_single_relationship
        UNWIND node_single_relationship.properties as node_relationship_property
        CREATE (diff_relationship_element)-[:DIFF_HAS_PROPERTY]->(diff_relationship_property:DiffProperty)
        SET diff_relationship_property = node_relationship_property.node_properties
        // -------------------------
        // add conflict for this relationship element
        // -------------------------
        FOREACH (i in CASE WHEN node_relationship_property.conflict_params IS NOT NULL THEN [1] ELSE [] END |
            CREATE (diff_relationship_property)-[:DIFF_HAS_CONFLICT]->(diff_relationship_property_conflict:DiffConflict)
            SET diff_relationship_property_conflict = node_relationship_property.conflict_params
        )
    }
}
        """
        self.add_to_query(query)

    def _build_conflict_params(self, enriched_conflict: EnrichedDiffConflict) -> dict[str, Any]:
        return {
            "uuid": enriched_conflict.uuid,
            "base_branch_action": enriched_conflict.base_branch_action.value,
            "base_branch_value": enriched_conflict.base_branch_value,
            "base_branch_changed_at": enriched_conflict.base_branch_changed_at.to_string()
            if enriched_conflict.base_branch_changed_at
            else None,
            "base_branch_label": enriched_conflict.base_branch_label,
            "diff_branch_action": enriched_conflict.diff_branch_action.value,
            "diff_branch_value": enriched_conflict.diff_branch_value,
            "diff_branch_changed_at": enriched_conflict.diff_branch_changed_at.to_string()
            if enriched_conflict.diff_branch_changed_at
            else None,
            "diff_branch_label": enriched_conflict.diff_branch_label,
            "selected_branch": enriched_conflict.selected_branch.value if enriched_conflict.selected_branch else None,
            "resolvable": enriched_conflict.resolvable,
        }

    def _build_diff_property_params(self, enriched_property: EnrichedDiffProperty) -> dict[str, Any]:
        conflict_params = None
        if enriched_property.conflict:
            conflict_params = self._build_conflict_params(enriched_conflict=enriched_property.conflict)
        return {
            "node_properties": {
                "property_type": enriched_property.property_type.value,
                "changed_at": enriched_property.changed_at.to_string(),
                "previous_value": enriched_property.previous_value,
                "new_value": enriched_property.new_value,
                "previous_label": enriched_property.previous_label,
                "new_label": enriched_property.new_label,
                "action": enriched_property.action,
                "path_identifier": enriched_property.path_identifier,
            },
            "conflict_params": conflict_params,
        }

    def _build_diff_attribute_params(self, enriched_attribute: EnrichedDiffAttribute) -> dict[str, Any]:
        property_props = [
            self._build_diff_property_params(enriched_property=prop) for prop in enriched_attribute.properties
        ]
        return {
            "node_properties": {
                "name": enriched_attribute.name,
                "changed_at": enriched_attribute.changed_at.to_string(),
                "action": enriched_attribute.action.value,
                "path_identifier": enriched_attribute.path_identifier,
                "num_added": enriched_attribute.num_added,
                "num_updated": enriched_attribute.num_updated,
                "num_removed": enriched_attribute.num_removed,
                "num_conflicts": enriched_attribute.num_conflicts,
                "contains_conflict": enriched_attribute.contains_conflict,
            },
            "properties": property_props,
        }

    def _build_diff_single_relationship_params(
        self, enriched_single_relationship: EnrichedDiffSingleRelationship
    ) -> dict[str, Any]:
        property_props = [
            self._build_diff_property_params(enriched_property=prop) for prop in enriched_single_relationship.properties
        ]
        conflict_params = None
        if enriched_single_relationship.conflict:
            conflict_params = self._build_conflict_params(enriched_conflict=enriched_single_relationship.conflict)
        return {
            "node_properties": {
                "changed_at": enriched_single_relationship.changed_at.to_string(),
                "action": enriched_single_relationship.action,
                "peer_id": enriched_single_relationship.peer_id,
                "peer_label": enriched_single_relationship.peer_label,
                "path_identifier": enriched_single_relationship.path_identifier,
                "num_added": enriched_single_relationship.num_added,
                "num_updated": enriched_single_relationship.num_updated,
                "num_removed": enriched_single_relationship.num_removed,
                "num_conflicts": enriched_single_relationship.num_conflicts,
                "contains_conflict": enriched_single_relationship.contains_conflict,
            },
            "conflict_params": conflict_params,
            "properties": property_props,
        }

    def _build_diff_relationship_params(self, enriched_relationship: EnrichedDiffRelationship) -> dict[str, Any]:
        single_relationship_props = [
            self._build_diff_single_relationship_params(enriched_single_relationship=esr)
            for esr in enriched_relationship.relationships
        ]
        return {
            "node_properties": {
                "name": enriched_relationship.name,
                "identifier": enriched_relationship.identifier,
                "label": enriched_relationship.label,
                "cardinality": enriched_relationship.cardinality.value,
                "changed_at": enriched_relationship.changed_at.to_string()
                if enriched_relationship.changed_at
                else None,
                "action": enriched_relationship.action,
                "path_identifier": enriched_relationship.path_identifier,
                "num_added": enriched_relationship.num_added,
                "num_updated": enriched_relationship.num_updated,
                "num_removed": enriched_relationship.num_removed,
                "num_conflicts": enriched_relationship.num_conflicts,
                "contains_conflict": enriched_relationship.contains_conflict,
            },
            "relationships": single_relationship_props,
        }

    def _build_diff_node_params(self, enriched_node: EnrichedDiffNode) -> dict[str, Any]:
        attribute_props = [
            self._build_diff_attribute_params(enriched_attribute=attribute) for attribute in enriched_node.attributes
        ]
        relationship_props = [
            self._build_diff_relationship_params(relationship) for relationship in enriched_node.relationships
        ]
        conflict_params = None
        if enriched_node.conflict:
            conflict_params = self._build_conflict_params(enriched_conflict=enriched_node.conflict)
        return {
            "node_properties": {
                "uuid": enriched_node.uuid,
                "kind": enriched_node.kind,
                "label": enriched_node.label,
                "changed_at": enriched_node.changed_at.to_string() if enriched_node.changed_at else None,
                "action": enriched_node.action.value,
                "path_identifier": enriched_node.path_identifier,
                "num_added": enriched_node.num_added,
                "num_updated": enriched_node.num_updated,
                "num_removed": enriched_node.num_removed,
                "num_conflicts": enriched_node.num_conflicts,
                "contains_conflict": enriched_node.contains_conflict,
            },
            "conflict_params": conflict_params,
            "attributes": attribute_props,
            "relationships": relationship_props,
        }

    def _build_node_batch_params(self) -> dict[str, list[dict[str, Any]]]:
        node_details: list[dict[str, Any]] = []
        for node_create_request in self.node_create_batch:
            node_details.append(
                {
                    "root_uuid": node_create_request.root_uuid,
                    "node_map": self._build_diff_node_params(enriched_node=node_create_request.node),
                }
            )
        return {"node_details_list": node_details}


class EnrichedNodesLinkQuery(Query):
    name = "enriched_nodes_link"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, enriched_diffs: EnrichedDiffs, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.enriched_diffs = enriched_diffs

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        parent_links_list = []
        for diff_root in (self.enriched_diffs.base_branch_diff, self.enriched_diffs.diff_branch_diff):
            for node in diff_root.nodes:
                parent_links_list.extend(self._build_node_parent_links(enriched_node=node, root_uuid=diff_root.uuid))
        self.params = {"node_links_list": parent_links_list}
        query = """
UNWIND $node_links_list AS node_link_details
WITH
    node_link_details.root_uuid AS root_uuid,
    node_link_details.parent_uuid AS parent_uuid,
    node_link_details.child_uuid AS child_uuid,
    node_link_details.relationship_name AS relationship_name
CALL {
    WITH root_uuid, parent_uuid, child_uuid, relationship_name
    MATCH (diff_root {uuid: root_uuid})
    MATCH (diff_root)-[:DIFF_HAS_NODE]->(parent_node:DiffNode {uuid: parent_uuid})
        -[:DIFF_HAS_RELATIONSHIP]->(diff_rel_group:DiffRelationship {name: relationship_name})
    MATCH (diff_root)-[:DIFF_HAS_NODE]->(child_node:DiffNode {uuid: child_uuid})
    MERGE (diff_rel_group)-[:DIFF_HAS_NODE]->(child_node)
}
        """
        self.add_to_query(query)

    def _build_node_parent_links(self, enriched_node: EnrichedDiffNode, root_uuid: str) -> list[dict[str, str]]:
        if not enriched_node.relationships:
            return []
        parent_links = []
        for relationship in enriched_node.relationships:
            for child_node in relationship.nodes:
                parent_links.append(
                    {
                        "parent_uuid": enriched_node.uuid,
                        "relationship_name": relationship.name,
                        "child_uuid": child_node.uuid,
                        "root_uuid": root_uuid,
                    }
                )
                parent_links.extend(self._build_node_parent_links(enriched_node=child_node, root_uuid=root_uuid))
        return parent_links



from pydantic import BaseModel, Field

from infrahub.core.constants import DiffAction
from infrahub.core.query.utils import filter_and, filter_or


class IncExclFilterOptions(BaseModel):
    includes: list[str] = Field(default_factory=list)
    excludes: list[str] = Field(default_factory=list)

    @property
    def is_empty(self) -> bool:
        if not self.includes and not self.excludes:
            return True
        return False


class IncExclActionFilterOptions(BaseModel):
    includes: set[DiffAction] = Field(default_factory=list)
    excludes: set[DiffAction] = Field(default_factory=list)

    @property
    def is_empty(self) -> bool:
        if not self.includes and not self.excludes:
            return True
        return False


class EnrichedDiffQueryFilters(BaseModel):
    ids: list[str] = Field(default_factory=list)
    kind: IncExclFilterOptions = IncExclFilterOptions()
    namespace: IncExclFilterOptions = IncExclFilterOptions()
    status: IncExclActionFilterOptions = IncExclActionFilterOptions()

    @property
    def is_empty(self) -> bool:
        if not self.ids and self.kind.is_empty and self.namespace.is_empty and self.status.is_empty:
            return True
        return False

    def generate(self) -> tuple[str, dict]:
        default_filter = ""

        params = {}

        if self.ids:
            params["ids"] = self.ids
            return "diff_node.uuid in $ids", params

        filters_list = []

        if self.is_empty:
            return default_filter, params

        # KIND, Pass the list directly
        if not self.kind.is_empty:
            filter_kind = []
            if self.kind.includes:
                filter_kind.append("diff_node.kind IN $filter_kind_includes")
                params["filter_kind_includes"] = self.kind.includes

            if self.kind.excludes:
                filter_kind.append("NOT(diff_node.kind IN $filter_kind_excludes)")
                params["filter_kind_excludes"] = self.kind.excludes

            filters_list.append(filter_and(filter_kind))

        # NAMESPACE, match on the start of the kind
        if not self.namespace.is_empty:
            filter_namespace = []
            if self.namespace.includes:
                filter_namespace.append(
                    filter_or([f'diff_node.kind STARTS WITH "{ns}"' for ns in self.namespace.includes])
                )
            if self.namespace.excludes:
                filter_namespace.append(
                    filter_and([f'NOT(diff_node.kind STARTS WITH "{ns}")' for ns in self.namespace.excludes])
                )
            filters_list.append(filter_and(filter_namespace))

        # STATUS, Pass the list directly
        if not self.status.is_empty:
            filter_action = []
            if self.status.includes:
                filter_action.append("diff_node.action IN $filter_status_includes")
                params["filter_status_includes"] = [str(item.value).lower() for item in self.status.includes]

            if self.status.excludes:
                filter_action.append("NOT(diff_node.action IN $filter_status_excludes)")
                params["filter_status_excludes"] = [str(item.value).lower() for item in self.status.excludes]

            filters_list.append(filter_and(filter_action))

        return filter_and(filters_list), params



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Optional, Union, get_args, get_origin
from uuid import UUID

import ujson
from infrahub_sdk.uuidt import UUIDT
from pydantic import BaseModel

from infrahub.core.constants import NULL_VALUE
from infrahub.core.query.standard_node import (
    StandardNodeCreateQuery,
    StandardNodeDeleteQuery,
    StandardNodeGetItemQuery,
    StandardNodeGetListQuery,
    StandardNodeQuery,
    StandardNodeUpdateQuery,
)
from infrahub.exceptions import Error, InitializationError

# pylint: disable=redefined-builtin

if TYPE_CHECKING:
    from neo4j.graph import Node as Neo4jNode
    from pydantic.fields import FieldInfo
    from typing_extensions import Self

    from infrahub.core.query import Query
    from infrahub.database import InfrahubDatabase


class StandardNode(BaseModel):
    id: Optional[str] = None
    uuid: Optional[UUID] = None

    _query: type[StandardNodeQuery] = StandardNodeCreateQuery
    _exclude_attrs: list[str] = ["id", "uuid", "_query"]

    @classmethod
    def get_type(cls) -> str:
        return cls.__name__

    def get_id(self) -> str:
        if not self.id:
            raise ValueError("id isn't defined yet")
        return self.id

    @staticmethod
    def guess_field_type(field: FieldInfo) -> Any:
        """Return the type of a Pydantic model field.

        Here we mimic a fraction of the old Pydantic type analysis.
        https://github.com/pydantic/pydantic/blob/2e939dc3bfb88f54efb66f8f1a031ff22e4f9865/pydantic/fields.py#L539%23L758
        """
        annotation_origin = get_origin(field.annotation)
        annotation_args = get_args(field.annotation)

        if (annotation_origin == Union and len(annotation_args) == 2 and type(None) in annotation_args) or (
            annotation_origin is list and len(annotation_args)
        ):
            return get_origin(annotation_args[0]) or annotation_args[0]

        return annotation_origin or field.annotation

    def get_uuid(self) -> UUID:
        if self.uuid:
            return self.uuid

        raise InitializationError("The root node has not been initialized with a uuid")

    async def to_graphql(self, fields: dict) -> dict:
        response: dict[str, Any] = {"id": self.uuid}

        for field_name in fields.keys():
            if field_name in ["id"]:
                continue
            if field_name == "__typename":
                response[field_name] = self.get_type()
                continue
            field = getattr(self, field_name)
            if field is None:
                response[field_name] = None
                continue
            response[field_name] = field

        return response

    async def save(self, db: InfrahubDatabase) -> bool:
        """Create or Update the Node in the database."""

        if self.id:
            return await self.update(db=db)

        return await self.create(db=db)

    async def delete(self, db: InfrahubDatabase) -> None:
        """Delete the Node in the database."""

        query: Query = await StandardNodeDeleteQuery.init(db=db, node=self)
        await query.execute(db=db)

    async def create(self, db: InfrahubDatabase) -> bool:
        """Create a new node in the database."""

        query: Query = await self._query.init(db=db, node=self)
        await query.execute(db=db)

        result = query.get_result()
        if not result:
            raise Error(f"Unable to create the node {self.get_type()}")
        node = result.get("n")

        self.id = node.element_id
        self.uuid = node["uuid"]

        return True

    async def update(self, db: InfrahubDatabase) -> bool:
        """Update the node in the database if needed."""

        query: Query = await StandardNodeUpdateQuery.init(db=db, node=self)
        await query.execute(db=db)
        result = query.get_result()

        if not result:
            raise Error(f"Unexpected error, unable to update the node {self.id} / {self.uuid}.")

        return True

    @classmethod
    async def get(cls, id: str, db: InfrahubDatabase) -> Optional[Self]:
        """Get a node from the database identified by its ID."""

        node = await cls._get_item_raw(id=id, db=db)
        if node:
            return cls.from_db(node)

        return None

    @classmethod
    async def _get_item_raw(cls, id: str, db: InfrahubDatabase) -> Neo4jNode:
        query: Query = await StandardNodeGetItemQuery.init(db=db, node_id=id, node_type=cls.get_type())
        await query.execute(db=db)

        result = query.get_result()
        if not result:
            return None

        return result.get("n")

    @classmethod
    def from_db(cls, node: Neo4jNode, extras: Optional[dict[str, Any]] = None) -> Self:
        """Convert a Neo4j Node to a Infrahub StandardNode

        Args:
            node (neo4j.graph.Node): Neo4j Node object

        Returns:
            StandardNode: Proper StandardNode object
        """

        attrs = {}
        node_data = dict(node)
        extras = extras or {}
        node_data.update(extras)
        attrs["id"] = node.element_id
        for key, value in node_data.items():
            if key not in cls.model_fields:
                continue

            field_type = cls.guess_field_type(cls.model_fields[key])

            if value == NULL_VALUE:
                attrs[key] = None
            elif issubclass(field_type, (int, float, bool, str, UUID)):
                attrs[key] = value
            elif isinstance(value, (str, bytes)):
                attrs[key] = ujson.loads(value)

        return cls(**attrs)

    def to_db(self) -> dict[str, Any]:
        data = {}

        if not self.uuid:
            data["uuid"] = str(UUIDT())
        else:
            data["uuid"] = str(self.uuid)

        for attr_name, field in self.model_fields.items():
            if attr_name in self._exclude_attrs:
                continue

            attr_value = getattr(self, attr_name)
            field_type = self.guess_field_type(field)

            if attr_value is None:
                data[attr_name] = NULL_VALUE
            elif inspect.isclass(field_type) and issubclass(field_type, BaseModel):
                if isinstance(attr_value, list):
                    clean_value = [item.model_dump() for item in attr_value]
                    data[attr_name] = ujson.dumps(clean_value)
                else:
                    data[attr_name] = attr_value.model_dump_json()
            elif issubclass(field_type, (int, float, bool, str, UUID)):
                data[attr_name] = attr_value
            else:
                data[attr_name] = ujson.dumps(attr_value)

        return data

    @classmethod
    async def get_list(
        cls, db: InfrahubDatabase, limit: int = 1000, ids: list[str] | None = None, name: str | None = None, **kwargs
    ) -> list[Self]:
        query: Query = await StandardNodeGetListQuery.init(
            db=db, node_class=cls, ids=ids, node_name=name, limit=limit, **kwargs
        )
        await query.execute(db=db)

        return [cls.from_db(result.get("n")) for result in query.get_results()]



from typing import Any

from infrahub.core.constants import DiffAction
from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import NodeDiffFieldSummary, TrackingId


class EnrichedDiffNodeFieldSummaryQuery(Query):
    """
    Get node kind and names of all altered attributes and relationships for each kind
    """

    name = "enriched_diff_node_field_summary"
    type = QueryType.READ

    def __init__(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.diff_branch_name = diff_branch_name
        self.tracking_id = tracking_id
        self.diff_id = diff_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        if self.tracking_id is None and self.diff_id is None:
            raise RuntimeError("Either tacking_id or diff_id is required")
        self.params = {
            "unchanged_str": DiffAction.UNCHANGED.value,
            "diff_branch": self.diff_branch_name,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
            "diff_id": self.diff_id,
        }
        query = """
        MATCH (diff_root:DiffRoot)
        WHERE diff_root.diff_branch = $diff_branch
        AND (diff_root.tracking_id = $tracking_id OR $tracking_id IS NULL)
        AND (diff_root.uuid = $diff_id OR $diff_id IS NULL)
        OPTIONAL MATCH (diff_root)-[:DIFF_HAS_NODE]->(n:DiffNode)
        WHERE n.action <> $unchanged_str
        WITH DISTINCT n.kind AS kind
        CALL {
            WITH kind
            OPTIONAL MATCH (n:DiffNode {kind: kind})-[:DIFF_HAS_ATTRIBUTE]->(a:DiffAttribute)
            WHERE n.action <> $unchanged_str
            AND a.action <> $unchanged_str
            WITH DISTINCT a.name AS attr_name
            RETURN collect(attr_name) AS attr_names
        }
        WITH kind, attr_names
        CALL {
            WITH kind
            OPTIONAL MATCH (n:DiffNode {kind: kind})-[:DIFF_HAS_RELATIONSHIP]->(r:DiffRelationship)
            WHERE n.action <> $unchanged_str
            AND r.action <> $unchanged_str
            WITH DISTINCT r.name AS rel_name
            RETURN collect(rel_name) AS rel_names
        }
        """
        self.add_to_query(query=query)
        self.return_labels = ["kind", "attr_names", "rel_names"]

    async def get_field_summaries(self) -> list[NodeDiffFieldSummary]:
        field_summaries = []
        for result in self.get_results():
            kind = result.get_as_type(label="kind", return_type=str)
            attr_names = result.get_as_type(label="attr_names", return_type=list[str])
            rel_names = result.get_as_type(label="rel_names", return_type=list[str])
            if attr_names or rel_names:
                field_summaries.append(
                    NodeDiffFieldSummary(kind=kind, attribute_names=set(attr_names), relationship_names=set(rel_names))
                )
        return field_summaries
